# -*- coding: utf-8 -*-
from django import template

register = template.Library()


class IndentByNode(template.Node):
    def __init__(self, nodelist, indent_level, if_statement):
        self.nodelist = nodelist
        self.indent_level = template.Variable(indent_level)
        if if_statement:
            self.if_statement = template.Variable(if_statement)
        else:
            self.if_statement = None

    def render(self, context):
        indent_level = self.indent_level.resolve(context)
        if self.if_statement:
            try:
                if_statement = bool(self.if_statement.resolve(context))
            except template.VariableDoesNotExist:
                if_statement = False
        else:
            if_statement = True
        output = self.nodelist.render(context)
        if if_statement:
            indent = " " * indent_level
            output = indent + indent.join(output.splitlines(True))
        return output


@register.tag
def indentby(parser, token):
    """
    Add indentation to text between the tags by the given indentation level.

    {% indentby <indent_level> [if <statement>] %}
    ...
    {% endindentby %}

    Arguments:
      indent_level - Number of spaces to indent text with.
      statement - Only apply indent_level if the boolean statement evalutates to True.
    """
    args = token.split_contents()
    largs = len(args)
    if largs not in (2, 4):
        raise template.TemplateSyntaxError("indentby tag requires 1 or 3 arguments")
    indent_level = args[1]
    if_statement = None
    if largs == 4:
        if_statement = args[3]
    nodelist = parser.parse(("endindentby",))
    parser.delete_first_token()
    return IndentByNode(nodelist, indent_level, if_statement)



# -*- coding: utf-8 -*-
"""
MongoDB model fields emulating Django Extensions' additional model fields

These fields are essentially identical to existing Extensions fields.
"""

import re
import datetime
from django import forms
from django.db.models.constants import LOOKUP_SEP
from django.template.defaultfilters import slugify
from django.utils.translation import gettext_lazy as _
from mongoengine.fields import StringField, DateTimeField

import uuid


class SlugField(StringField):
    description = _("String (up to %(max_length)s)")

    def __init__(self, *args, **kwargs):
        kwargs["max_length"] = kwargs.get("max_length", 50)
        # Set db_index=True unless it's been set manually.
        if "db_index" not in kwargs:
            kwargs["db_index"] = True
        super().__init__(*args, **kwargs)

    def get_internal_type(self):
        return "SlugField"

    def formfield(self, **kwargs):
        defaults = {"form_class": forms.SlugField}
        defaults.update(kwargs)
        return super().formfield(**defaults)


class AutoSlugField(SlugField):
    """
    AutoSlugField, adapted for MongoDB

    By default, sets editable=False, blank=True.

    Required arguments:

    populate_from
        Specifies which field or list of fields the slug is populated from.

    Optional arguments:

    separator
        Defines the used separator (default: '-')

    overwrite
        If set to True, overwrites the slug on every save (default: False)

    Inspired by SmileyChris' Unique Slugify snippet:
    https://www.djangosnippets.org/snippets/690/
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        kwargs.setdefault("editable", False)

        populate_from = kwargs.pop("populate_from", None)
        if populate_from is None:
            raise ValueError("missing 'populate_from' argument")
        else:
            self._populate_from = populate_from

        self.slugify_function = kwargs.pop("slugify_function", slugify)
        self.separator = kwargs.pop("separator", str("-"))
        self.overwrite = kwargs.pop("overwrite", False)
        super().__init__(*args, **kwargs)

    def _slug_strip(self, value):
        """
        Clean up a slug by removing slug separator characters that occur at
        the beginning or end of a slug.

        If an alternate separator is used, it will also replace any instances
        of the default '-' separator with the new separator.
        """
        re_sep = "(?:-|%s)" % re.escape(self.separator)
        value = re.sub("%s+" % re_sep, self.separator, value)
        return re.sub(r"^%s+|%s+$" % (re_sep, re_sep), "", value)

    def slugify_func(self, content):
        return self.slugify_function(content)

    def create_slug(self, model_instance, add):
        # get fields to populate from and slug field to set
        if not isinstance(self._populate_from, (list, tuple)):
            self._populate_from = (self._populate_from,)
        slug_field = model_instance._meta.get_field(self.attname)

        if add or self.overwrite:
            # slugify the original field content and set next step to 2
            slug_for_field = lambda lookup_value: self.slugify_func(
                self.get_slug_fields(model_instance, lookup_value)
            )
            slug = self.separator.join(map(slug_for_field, self._populate_from))
            next = 2
        else:
            # get slug from the current model instance and calculate next
            # step from its number, clean-up
            slug = self._slug_strip(getattr(model_instance, self.attname))
            next = slug.split(self.separator)[-1]
            if next.isdigit():
                slug = self.separator.join(slug.split(self.separator)[:-1])
                next = int(next)
            else:
                next = 2

        # strip slug depending on max_length attribute of the slug field
        # and clean-up
        slug_len = slug_field.max_length
        if slug_len:
            slug = slug[:slug_len]
        slug = self._slug_strip(slug)
        original_slug = slug

        # exclude the current model instance from the queryset used in finding
        # the next valid slug
        queryset = model_instance.__class__._default_manager.all()
        if model_instance.pk:
            queryset = queryset.exclude(pk=model_instance.pk)

        # form a kwarg dict used to impliment any unique_together constraints
        kwargs = {}
        for params in model_instance._meta.unique_together:
            if self.attname in params:
                for param in params:
                    kwargs[param] = getattr(model_instance, param, None)
        kwargs[self.attname] = slug

        # increases the number while searching for the next valid slug
        # depending on the given slug, clean-up
        while not slug or queryset.filter(**kwargs):
            slug = original_slug
            end = "%s%s" % (self.separator, next)
            end_len = len(end)
            if slug_len and len(slug) + end_len > slug_len:
                slug = slug[: slug_len - end_len]
                slug = self._slug_strip(slug)
            slug = "%s%s" % (slug, end)
            kwargs[self.attname] = slug
            next += 1
        return slug

    def get_slug_fields(self, model_instance, lookup_value):
        lookup_value_path = lookup_value.split(LOOKUP_SEP)
        attr = model_instance
        for elem in lookup_value_path:
            try:
                attr = getattr(attr, elem)
            except AttributeError:
                raise AttributeError(
                    "value {} in AutoSlugField's 'populate_from' argument {} returned an error - {} has no attribute {}".format(  # noqa: E501
                        elem, lookup_value, attr, elem
                    )
                )

        if callable(attr):
            return "%s" % attr()

        return attr

    def pre_save(self, model_instance, add):
        value = str(self.create_slug(model_instance, add))
        setattr(model_instance, self.attname, value)
        return value

    def get_internal_type(self):
        return "SlugField"


class CreationDateTimeField(DateTimeField):
    """
    CreationDateTimeField

    By default, sets editable=False, blank=True, default=datetime.now
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("default", datetime.datetime.now)
        DateTimeField.__init__(self, *args, **kwargs)

    def get_internal_type(self):
        return "DateTimeField"


class ModificationDateTimeField(CreationDateTimeField):
    """
    ModificationDateTimeField

    By default, sets editable=False, blank=True, default=datetime.now

    Sets value to datetime.now() on each save of the model.
    """

    def pre_save(self, model, add):
        value = datetime.datetime.now()
        setattr(model, self.attname, value)
        return value

    def get_internal_type(self):
        return "DateTimeField"


class UUIDVersionError(Exception):
    pass


class UUIDField(StringField):
    """
    UUIDField

    By default uses UUID version 1 (generate from host ID, sequence number and current time)

    The field support all uuid versions which are natively supported by the uuid python module.
    For more information see: https://docs.python.org/lib/module-uuid.html
    """  # noqa: E501

    def __init__(
        self,
        verbose_name=None,
        name=None,
        auto=True,
        version=1,
        node=None,
        clock_seq=None,
        namespace=None,
        **kwargs,
    ):
        kwargs["max_length"] = 36
        self.auto = auto
        self.version = version
        if version == 1:
            self.node, self.clock_seq = node, clock_seq
        elif version == 3 or version == 5:
            self.namespace, self.name = namespace, name
        StringField.__init__(self, verbose_name, name, **kwargs)

    def get_internal_type(self):
        return StringField.__name__

    def contribute_to_class(self, cls, name):
        if self.primary_key:
            assert not cls._meta.has_auto_field, (
                "A model can't have more than one AutoField: %s %s %s; have %s"
                % (self, cls, name, cls._meta.auto_field)
            )
            super().contribute_to_class(cls, name)
            cls._meta.has_auto_field = True
            cls._meta.auto_field = self
        else:
            super().contribute_to_class(cls, name)

    def create_uuid(self):
        if not self.version or self.version == 4:
            return uuid.uuid4()
        elif self.version == 1:
            return uuid.uuid1(self.node, self.clock_seq)
        elif self.version == 2:
            raise UUIDVersionError("UUID version 2 is not supported.")
        elif self.version == 3:
            return uuid.uuid3(self.namespace, self.name)
        elif self.version == 5:
            return uuid.uuid5(self.namespace, self.name)
        else:
            raise UUIDVersionError("UUID version %s is not valid." % self.version)

    def pre_save(self, model_instance, add):
        if self.auto and add:
            value = str(self.create_uuid())
            setattr(model_instance, self.attname, value)
            return value
        else:
            value = super().pre_save(model_instance, add)
            if self.auto and not value:
                value = str(self.create_uuid())
                setattr(model_instance, self.attname, value)
        return value



# -*- coding: utf-8 -*-
import configparser


def parse_mysql_cnf(dbinfo):
    """
    Attempt to parse mysql database config file for connection settings.
    Ideally we would hook into django's code to do this, but read_default_file is
    handled by the mysql C libs so we have to emulate the behaviour

    Settings that are missing will return ''
    returns (user, password, database_name, database_host, database_port)
    """
    read_default_file = dbinfo.get("OPTIONS", {}).get("read_default_file")
    if read_default_file:
        config = configparser.RawConfigParser(
            {
                "user": "",
                "password": "",
                "database": "",
                "host": "",
                "port": "",
                "socket": "",
            }
        )
        import os

        config.read(os.path.expanduser(read_default_file))
        try:
            user = config.get("client", "user")
            password = config.get("client", "password")
            database_name = config.get("client", "database")
            database_host = config.get("client", "host")
            database_port = config.get("client", "port")
            socket = config.get("client", "socket")

            if database_host == "localhost" and socket:
                # mysql actually uses a socket if host is localhost
                database_host = socket

            return user, password, database_name, database_host, database_port

        except configparser.NoSectionError:
            pass

    return "", "", "", "", ""



# -*- coding: utf-8 -*-

import functools
import json
import re

from django.conf import settings
from django.contrib.admindocs.views import simplify_regex
from django.core.exceptions import ViewDoesNotExist
from django.core.management.base import BaseCommand, CommandError
from django.utils import translation

from django_extensions.management.color import color_style, no_style
from django_extensions.management.utils import signalcommand

from django.urls import URLPattern, URLResolver  # type: ignore


class RegexURLPattern:  # type: ignore
    pass


class RegexURLResolver:  # type: ignore
    pass


class LocaleRegexURLResolver:  # type: ignore
    pass


def describe_pattern(p):
    return str(p.pattern)


FMTR = {
    "dense": "{url}\t{module}\t{url_name}\t{decorator}",
    "table": "{url},{module},{url_name},{decorator}",
    "aligned": "{url},{module},{url_name},{decorator}",
    "verbose": "{url}\n\tController: {module}\n\tURL Name: {url_name}\n\tDecorators: {decorator}\n",  # noqa: E501
    "json": "",
    "pretty-json": "",
}


class Command(BaseCommand):
    help = "Displays all of the url matching routes for the project."

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--unsorted",
            "-u",
            action="store_true",
            dest="unsorted",
            help="Show urls unsorted but same order as found in url patterns",
        )
        parser.add_argument(
            "--language",
            "-l",
            dest="language",
            help="Only show this language code (useful for i18n_patterns)",
        )
        parser.add_argument(
            "--decorator",
            "-d",
            action="append",
            dest="decorator",
            default=[],
            help="Show the presence of given decorator on views",
        )
        parser.add_argument(
            "--format",
            "-f",
            dest="format_style",
            default="dense",
            help="Style of the output. Choices: %s" % FMTR.keys(),
        )
        parser.add_argument(
            "--urlconf",
            "-c",
            dest="urlconf",
            default="ROOT_URLCONF",
            help="Set the settings URL conf variable to use",
        )

    @signalcommand
    def handle(self, *args, **options):
        style = no_style() if options["no_color"] else color_style()

        language = options["language"]
        if language is not None:
            translation.activate(language)
            self.LANGUAGES = [
                (code, name)
                for code, name in getattr(settings, "LANGUAGES", [])
                if code == language
            ]
        else:
            self.LANGUAGES = getattr(settings, "LANGUAGES", ((None, None),))

        decorator = options["decorator"]
        if not decorator:
            decorator = ["login_required"]

        format_style = options["format_style"]
        if format_style not in FMTR:
            raise CommandError(
                "Format style '%s' does not exist. Options: %s"
                % (
                    format_style,
                    ", ".join(sorted(FMTR.keys())),
                )
            )
        pretty_json = format_style == "pretty-json"
        if pretty_json:
            format_style = "json"
        fmtr = FMTR[format_style]

        urlconf = options["urlconf"]

        views = []
        if not hasattr(settings, urlconf):
            raise CommandError(
                "Settings module {} does not have the attribute {}.".format(
                    settings, urlconf
                )
            )

        try:
            urlconf = __import__(getattr(settings, urlconf), {}, {}, [""])
        except Exception as e:
            if options["traceback"]:
                import traceback

                traceback.print_exc()
            raise CommandError(
                "Error occurred while trying to load %s: %s"
                % (getattr(settings, urlconf), str(e))
            )

        view_functions = self.extract_views_from_urlpatterns(urlconf.urlpatterns)
        for func, regex, url_name in view_functions:
            if hasattr(func, "__globals__"):
                func_globals = func.__globals__
            elif hasattr(func, "func_globals"):
                func_globals = func.func_globals
            else:
                func_globals = {}

            decorators = [d for d in decorator if d in func_globals]

            if isinstance(func, functools.partial):
                func = func.func
                decorators.insert(0, "functools.partial")

            if hasattr(func, "view_class"):
                func = func.view_class
            if hasattr(func, "__name__"):
                func_name = func.__name__
            elif hasattr(func, "__class__"):
                func_name = "%s()" % func.__class__.__name__
            else:
                func_name = re.sub(r" at 0x[0-9a-f]+", "", repr(func))

            module = "{0}.{1}".format(func.__module__, func_name)
            url_name = url_name or ""
            url = simplify_regex(regex)
            decorator = ", ".join(decorators)

            if format_style == "json":
                views.append(
                    {
                        "url": url,
                        "module": module,
                        "name": url_name,
                        "decorators": decorator,
                    }
                )
            else:
                views.append(
                    fmtr.format(
                        module="{0}.{1}".format(
                            style.MODULE(func.__module__), style.MODULE_NAME(func_name)
                        ),
                        url_name=style.URL_NAME(url_name),
                        url=style.URL(url),
                        decorator=decorator,
                    ).strip()
                )

        if not options["unsorted"] and format_style != "json":
            views = sorted(views)

        if format_style == "aligned":
            views = [row.split(",", 3) for row in views]
            widths = [len(max(columns, key=len)) for columns in zip(*views)]
            views = [
                "   ".join(
                    "{0:<{1}}".format(cdata, width) for width, cdata in zip(widths, row)
                )
                for row in views
            ]
        elif format_style == "table":
            # Reformat all data and show in a table format

            views = [row.split(",", 3) for row in views]
            widths = [len(max(columns, key=len)) for columns in zip(*views)]
            table_views = []

            header = (
                style.MODULE_NAME("URL"),
                style.MODULE_NAME("Module"),
                style.MODULE_NAME("Name"),
                style.MODULE_NAME("Decorator"),
            )
            table_views.append(
                " | ".join(
                    "{0:<{1}}".format(title, width)
                    for width, title in zip(widths, header)
                )
            )
            table_views.append("-+-".join("-" * width for width in widths))

            for row in views:
                table_views.append(
                    " | ".join(
                        "{0:<{1}}".format(cdata, width)
                        for width, cdata in zip(widths, row)
                    )
                )

            # Replace original views so we can return the same object
            views = table_views

        elif format_style == "json":
            if pretty_json:
                return json.dumps(views, indent=4)
            return json.dumps(views)

        return "\n".join([v for v in views]) + "\n"

    def extract_views_from_urlpatterns(self, urlpatterns, base="", namespace=None):
        """
        Return a list of views from a list of urlpatterns.

        Each object in the returned list is a three-tuple: (view_func, regex, name)
        """
        views = []
        for p in urlpatterns:
            if isinstance(p, (URLPattern, RegexURLPattern)):
                try:
                    if not p.name:
                        name = p.name
                    elif namespace:
                        name = "{0}:{1}".format(namespace, p.name)
                    else:
                        name = p.name
                    pattern = describe_pattern(p)
                    views.append((p.callback, base + pattern, name))
                except ViewDoesNotExist:
                    continue
            elif isinstance(p, (URLResolver, RegexURLResolver)):
                try:
                    patterns = p.url_patterns
                except ImportError:
                    continue
                if namespace and p.namespace:
                    _namespace = "{0}:{1}".format(namespace, p.namespace)
                else:
                    _namespace = p.namespace or namespace
                pattern = describe_pattern(p)
                if isinstance(p, LocaleRegexURLResolver):
                    for language in self.LANGUAGES:
                        with translation.override(language[0]):
                            views.extend(
                                self.extract_views_from_urlpatterns(
                                    patterns, base + pattern, namespace=_namespace
                                )
                            )
                else:
                    views.extend(
                        self.extract_views_from_urlpatterns(
                            patterns, base + pattern, namespace=_namespace
                        )
                    )
            elif hasattr(p, "_get_callback"):
                try:
                    views.append(
                        (p._get_callback(), base + describe_pattern(p), p.name)
                    )
                except ViewDoesNotExist:
                    continue
            elif hasattr(p, "url_patterns") or hasattr(p, "_get_url_patterns"):
                try:
                    patterns = p.url_patterns
                except ImportError:
                    continue
                views.extend(
                    self.extract_views_from_urlpatterns(
                        patterns, base + describe_pattern(p), namespace=namespace
                    )
                )
            else:
                raise TypeError("%s does not appear to be a urlpattern object" % p)
        return views



# -*- coding: utf-8 -*-
"""
sqldiff.py - Prints the (approximated) difference between models and database

TODO:
 - better support for relations
 - better support for constraints (mainly postgresql?)
 - support for table spaces with postgresql
 - when a table is not managed (meta.managed==False) then only do a one-way
   sqldiff ? show differences from db->table but not the other way around since
   it's not managed.

KNOWN ISSUES:
 - MySQL has by far the most problems with introspection. Please be
   carefull when using MySQL with sqldiff.
   - Booleans are reported back as Integers, so there's no way to know if
     there was a real change.
   - Varchar sizes are reported back without unicode support so their size
     may change in comparison to the real length of the varchar.
   - Some of the 'fixes' to counter these problems might create false
     positives or false negatives.
"""

import importlib
import sys
import argparse
from typing import Dict, Union, Callable, Optional  # NOQA
from django.apps import apps
from django.core.management import BaseCommand, CommandError
from django.core.management.base import OutputWrapper
from django.core.management.color import no_style
from django.db import connection, transaction, models
from django.db.models import UniqueConstraint
from django.db.models.fields import AutoField, IntegerField
from django.db.models.options import normalize_together

from django_extensions.management.utils import signalcommand

ORDERING_FIELD: IntegerField = IntegerField("_order", null=True)


def flatten(lst, ltypes=(list, tuple)):
    ltype = type(lst)
    lst = list(lst)
    i = 0
    while i < len(lst):
        while isinstance(lst[i], ltypes):
            if not lst[i]:
                lst.pop(i)
                i -= 1
                break
            else:
                lst[i : i + 1] = lst[i]
        i += 1
    return ltype(lst)


def all_local_fields(meta):
    all_fields = []
    if meta.proxy:
        for parent in meta.parents:
            all_fields.extend(all_local_fields(parent._meta))
    else:
        for f in meta.local_fields:
            col_type = f.db_type(connection=connection)
            if col_type is None:
                continue
            all_fields.append(f)
    return all_fields


class SQLDiff:
    DATA_TYPES_REVERSE_OVERRIDE = {}  # type: Dict[int, Union[str, Callable]]

    IGNORE_MISSING_TABLES = [
        "django_migrations",
    ]

    DIFF_TYPES = [
        "error",
        "comment",
        "table-missing-in-db",
        "table-missing-in-model",
        "field-missing-in-db",
        "field-missing-in-model",
        "fkey-missing-in-db",
        "fkey-missing-in-model",
        "index-missing-in-db",
        "index-missing-in-model",
        "unique-missing-in-db",
        "unique-missing-in-model",
        "field-type-differ",
        "field-parameter-differ",
        "notnull-differ",
    ]
    DIFF_TEXTS = {
        "error": "error: %(0)s",
        "comment": "comment: %(0)s",
        "table-missing-in-db": "table '%(0)s' missing in database",
        "table-missing-in-model": "table '%(0)s' missing in models",
        "field-missing-in-db": "field '%(1)s' defined in model but missing in database",
        "field-missing-in-model": "field '%(1)s' defined in database but missing in model",  # noqa: E501
        "fkey-missing-in-db": "field '%(1)s' FOREIGN KEY defined in model but missing in database",  # noqa: E501
        "fkey-missing-in-model": "field '%(1)s' FOREIGN KEY defined in database but missing in model",  # noqa: E501
        "index-missing-in-db": "field '%(1)s' INDEX named '%(2)s' defined in model but missing in database",  # noqa: E501
        "index-missing-in-model": "field '%(1)s' INDEX defined in database schema but missing in model",  # noqa: E501
        "unique-missing-in-db": "field '%(1)s' UNIQUE named '%(2)s' defined in model but missing in database",  # noqa: E501
        "unique-missing-in-model": "field '%(1)s' UNIQUE defined in database schema but missing in model",  # noqa: E501
        "field-type-differ": "field '%(1)s' not of same type: db='%(3)s', model='%(2)s'",  # noqa: E501
        "field-parameter-differ": "field '%(1)s' parameters differ: db='%(3)s', model='%(2)s'",  # noqa: E501
        "notnull-differ": "field '%(1)s' null constraint should be '%(2)s' in the database",  # noqa: E501
    }

    SQL_FIELD_MISSING_IN_DB = lambda self, style, qn, args: "%s %s\n\t%s %s %s;" % (
        style.SQL_KEYWORD("ALTER TABLE"),
        style.SQL_TABLE(qn(args[0])),
        style.SQL_KEYWORD("ADD COLUMN"),
        style.SQL_FIELD(qn(args[1])),
        " ".join(
            style.SQL_COLTYPE(a) if i == 0 else style.SQL_KEYWORD(a)
            for i, a in enumerate(args[2:])
        ),
    )
    SQL_FIELD_MISSING_IN_MODEL = lambda self, style, qn, args: "%s %s\n\t%s %s;" % (
        style.SQL_KEYWORD("ALTER TABLE"),
        style.SQL_TABLE(qn(args[0])),
        style.SQL_KEYWORD("DROP COLUMN"),
        style.SQL_FIELD(qn(args[1])),
    )
    SQL_FKEY_MISSING_IN_DB = (
        lambda self, style, qn, args: "%s %s\n\t%s %s %s %s %s (%s)%s;"
        % (
            style.SQL_KEYWORD("ALTER TABLE"),
            style.SQL_TABLE(qn(args[0])),
            style.SQL_KEYWORD("ADD COLUMN"),
            style.SQL_FIELD(qn(args[1])),
            " ".join(
                style.SQL_COLTYPE(a) if i == 0 else style.SQL_KEYWORD(a)
                for i, a in enumerate(args[4:])
            ),
            style.SQL_KEYWORD("REFERENCES"),
            style.SQL_TABLE(qn(args[2])),
            style.SQL_FIELD(qn(args[3])),
            connection.ops.deferrable_sql(),
        )
    )
    SQL_INDEX_MISSING_IN_DB = lambda self, style, qn, args: "%s %s\n\t%s %s (%s%s);" % (
        style.SQL_KEYWORD("CREATE INDEX"),
        style.SQL_TABLE(qn(args[2])),
        # style.SQL_TABLE(qn("%s" % '_'.join('_'.join(a) if isinstance(a, (list, tuple)) else a for a in args[0:3] if a))),  # noqa: E501
        style.SQL_KEYWORD("ON"),
        style.SQL_TABLE(qn(args[0])),
        style.SQL_FIELD(", ".join(qn(e) for e in args[1])),
        style.SQL_KEYWORD(args[3]),
    )
    SQL_INDEX_MISSING_IN_MODEL = lambda self, style, qn, args: "%s %s;" % (
        style.SQL_KEYWORD("DROP INDEX"),
        style.SQL_TABLE(qn(args[1])),
    )
    SQL_UNIQUE_MISSING_IN_DB = (
        lambda self, style, qn, args: "%s %s\n\t%s %s %s (%s);"
        % (
            style.SQL_KEYWORD("ALTER TABLE"),
            style.SQL_TABLE(qn(args[0])),
            style.SQL_KEYWORD("ADD CONSTRAINT"),
            style.SQL_TABLE(qn(args[2])),
            style.SQL_KEYWORD("UNIQUE"),
            style.SQL_FIELD(", ".join(qn(e) for e in args[1])),
        )
    )
    SQL_UNIQUE_MISSING_IN_MODEL = lambda self, style, qn, args: "%s %s\n\t%s %s %s;" % (
        style.SQL_KEYWORD("ALTER TABLE"),
        style.SQL_TABLE(qn(args[0])),
        style.SQL_KEYWORD("DROP"),
        style.SQL_KEYWORD("CONSTRAINT"),
        style.SQL_TABLE(qn(args[1])),
    )
    SQL_FIELD_TYPE_DIFFER = lambda self, style, qn, args: "%s %s\n\t%s %s %s;" % (
        style.SQL_KEYWORD("ALTER TABLE"),
        style.SQL_TABLE(qn(args[0])),
        style.SQL_KEYWORD("MODIFY"),
        style.SQL_FIELD(qn(args[1])),
        style.SQL_COLTYPE(args[2]),
    )
    SQL_FIELD_PARAMETER_DIFFER = lambda self, style, qn, args: "%s %s\n\t%s %s %s;" % (
        style.SQL_KEYWORD("ALTER TABLE"),
        style.SQL_TABLE(qn(args[0])),
        style.SQL_KEYWORD("MODIFY"),
        style.SQL_FIELD(qn(args[1])),
        style.SQL_COLTYPE(args[2]),
    )
    SQL_NOTNULL_DIFFER = lambda self, style, qn, args: "%s %s\n\t%s %s %s %s;" % (
        style.SQL_KEYWORD("ALTER TABLE"),
        style.SQL_TABLE(qn(args[0])),
        style.SQL_KEYWORD("MODIFY"),
        style.SQL_FIELD(qn(args[1])),
        style.SQL_KEYWORD(args[2]),
        style.SQL_KEYWORD("NOT NULL"),
    )
    SQL_ERROR = lambda self, style, qn, args: style.NOTICE(
        "-- Error: %s" % style.ERROR(args[0])
    )
    SQL_COMMENT = lambda self, style, qn, args: style.NOTICE(
        "-- Comment: %s" % style.SQL_TABLE(args[0])
    )
    SQL_TABLE_MISSING_IN_DB = lambda self, style, qn, args: style.NOTICE(
        "-- Table missing: %s" % args[0]
    )
    SQL_TABLE_MISSING_IN_MODEL = lambda self, style, qn, args: style.NOTICE(
        "-- Model missing for table: %s" % args[0]
    )

    can_detect_notnull_differ = False
    can_detect_unsigned_differ = False
    unsigned_suffix = None  # type: Optional[str]

    def __init__(self, app_models, options, stdout, stderr):
        self.has_differences = None
        self.app_models = app_models
        self.options = options
        self.dense = options["dense_output"]
        self.stdout = stdout
        self.stderr = stderr

        self.introspection = connection.introspection

        self.differences = []
        self.unknown_db_fields = {}
        self.new_db_fields = set()
        self.null = {}
        self.unsigned = set()

        self.DIFF_SQL = {
            "error": self.SQL_ERROR,
            "comment": self.SQL_COMMENT,
            "table-missing-in-db": self.SQL_TABLE_MISSING_IN_DB,
            "table-missing-in-model": self.SQL_TABLE_MISSING_IN_MODEL,
            "field-missing-in-db": self.SQL_FIELD_MISSING_IN_DB,
            "field-missing-in-model": self.SQL_FIELD_MISSING_IN_MODEL,
            "fkey-missing-in-db": self.SQL_FKEY_MISSING_IN_DB,
            "fkey-missing-in-model": self.SQL_FIELD_MISSING_IN_MODEL,
            "index-missing-in-db": self.SQL_INDEX_MISSING_IN_DB,
            "index-missing-in-model": self.SQL_INDEX_MISSING_IN_MODEL,
            "unique-missing-in-db": self.SQL_UNIQUE_MISSING_IN_DB,
            "unique-missing-in-model": self.SQL_UNIQUE_MISSING_IN_MODEL,
            "field-type-differ": self.SQL_FIELD_TYPE_DIFFER,
            "field-parameter-differ": self.SQL_FIELD_PARAMETER_DIFFER,
            "notnull-differ": self.SQL_NOTNULL_DIFFER,
        }

    def load(self):
        self.cursor = connection.cursor()
        self.django_tables = self.introspection.django_table_names(
            only_existing=self.options["only_existing"]
        )
        # TODO: We are losing information about tables which are views here
        self.db_tables = [
            table_info.name
            for table_info in self.introspection.get_table_list(self.cursor)
        ]

        if self.can_detect_notnull_differ:
            self.load_null()

        if self.can_detect_unsigned_differ:
            self.load_unsigned()

    def load_null(self):
        raise NotImplementedError(
            (
                "load_null functions must be implemented if diff backend has "
                "'can_detect_notnull_differ' set to True"
            )
        )

    def load_unsigned(self):
        raise NotImplementedError(
            (
                "load_unsigned function must be implemented if diff backend has "
                "'can_detect_unsigned_differ' set to True"
            )
        )

    def add_app_model_marker(self, app_label, model_name):
        self.differences.append((app_label, model_name, []))

    def add_difference(self, diff_type, *args):
        assert diff_type in self.DIFF_TYPES, "Unknown difference type"
        self.differences[-1][-1].append((diff_type, args))

    def get_data_types_reverse_override(self):
        # type: () -> Dict[int, Union[str, Callable]]
        return self.DATA_TYPES_REVERSE_OVERRIDE

    def format_field_names(self, field_names):
        return field_names

    def sql_to_dict(self, query, param):
        """
        Execute query and return a dict

        sql_to_dict(query, param) -> list of dicts

        code from snippet at https://www.djangosnippets.org/snippets/1383/
        """
        cursor = connection.cursor()
        cursor.execute(query, param)
        fieldnames = [name[0] for name in cursor.description]
        fieldnames = self.format_field_names(fieldnames)
        result = []
        for row in cursor.fetchall():
            rowset = []
            for field in zip(fieldnames, row):
                rowset.append(field)
            result.append(dict(rowset))
        return result

    def get_field_model_type(self, field):
        return field.db_type(connection=connection)

    def get_field_db_type_kwargs(
        self,
        current_kwargs,
        description,
        field=None,
        table_name=None,
        reverse_type=None,
    ):
        return {}

    def get_field_db_type(self, description, field=None, table_name=None):
        # DB-API cursor.description
        #   (name, type_code, display_size, internal_size, precision, scale, null_ok)
        type_code = description[1]
        DATA_TYPES_REVERSE_OVERRIDE = self.get_data_types_reverse_override()
        if type_code in DATA_TYPES_REVERSE_OVERRIDE:
            reverse_type = DATA_TYPES_REVERSE_OVERRIDE[type_code]
        else:
            try:
                reverse_type = self.introspection.get_field_type(type_code, description)
            except KeyError:
                reverse_type = self.get_field_db_type_lookup(type_code)
                if not reverse_type:
                    # type_code not found in data_types_reverse map
                    key = (self.differences[-1][:2], description[:2])
                    if key not in self.unknown_db_fields:
                        self.unknown_db_fields[key] = 1
                        self.add_difference(
                            "comment",
                            "Unknown database type for field '%s' (%s)"
                            % (description[0], type_code),
                        )
                    return None

        if callable(reverse_type):
            reverse_type = reverse_type()

        kwargs = {}

        if isinstance(reverse_type, dict):
            kwargs.update(reverse_type["kwargs"])
            reverse_type = reverse_type["name"]

        if (
            type_code == 16946
            and field
            and getattr(field, "geom_type", None) == "POINT"
        ):
            reverse_type = "django.contrib.gis.db.models.fields.PointField"

        if isinstance(reverse_type, tuple):
            kwargs.update(reverse_type[1])
            reverse_type = reverse_type[0]

        if reverse_type == "CharField" and description[3]:
            kwargs["max_length"] = description[3]

        if reverse_type == "DecimalField":
            kwargs["max_digits"] = description[4]
            kwargs["decimal_places"] = (
                description[5] and abs(description[5]) or description[5]
            )

        if description[6]:
            kwargs["blank"] = True
            if reverse_type not in ("TextField", "CharField"):
                kwargs["null"] = True

        if field and getattr(field, "geography", False):
            kwargs["geography"] = True

        if reverse_type == "GeometryField":
            geo_col = description[0]
            # Getting a more specific field type and any additional parameters
            # from the `get_geometry_type` routine for the spatial backend.
            reverse_type, geo_params = self.introspection.get_geometry_type(
                table_name, geo_col
            )
            if geo_params:
                kwargs.update(geo_params)
            reverse_type = "django.contrib.gis.db.models.fields.%s" % reverse_type

        extra_kwargs = self.get_field_db_type_kwargs(
            kwargs, description, field, table_name, reverse_type
        )
        kwargs.update(extra_kwargs)

        field_class = self.get_field_class(reverse_type)
        field_db_type = field_class(**kwargs).db_type(connection=connection)

        tablespace = field.db_tablespace
        if not tablespace:
            tablespace = "public"
        if (
            tablespace,
            table_name,
            field.column,
        ) in self.unsigned and self.unsigned_suffix not in field_db_type:
            field_db_type = "%s %s" % (field_db_type, self.unsigned_suffix)

        return field_db_type

    def get_field_db_type_lookup(self, type_code):
        return None

    def get_field_class(self, class_path):
        if "." in class_path:
            module_path, package_name = class_path.rsplit(".", 1)
            module = importlib.import_module(module_path)
            return getattr(module, package_name)

        return getattr(models, class_path)

    def get_field_db_nullable(self, field, table_name):
        tablespace = field.db_tablespace
        if tablespace == "":
            tablespace = "public"
        attname = field.db_column or field.attname
        return self.null.get((tablespace, table_name, attname), "fixme")

    def strip_parameters(self, field_type):
        if field_type and field_type != "double precision":
            return field_type.split(" ")[0].split("(")[0].lower()
        return field_type

    def get_index_together(self, meta):
        indexes_normalized = []

        if hasattr(meta, "index_together"):
            # Django 4.2 deprecated index_together
            indexes_normalized += list(normalize_together(meta.index_together))

        for idx in meta.indexes:
            indexes_normalized.append(idx.fields)

        return self.expand_together(indexes_normalized, meta)

    def get_unique_together(self, meta):
        unique_normalized = list(normalize_together(meta.unique_together))

        for constraint in meta.constraints:
            if isinstance(constraint, UniqueConstraint):
                unique_normalized.append(constraint.fields)

        return self.expand_together(unique_normalized, meta)

    def expand_together(self, together, meta):
        new_together = []
        for fields in normalize_together(together):
            new_together.append(
                tuple(meta.get_field(field).attname for field in fields)
            )
        return new_together

    def find_unique_missing_in_db(
        self, meta, table_indexes, table_constraints, table_name, skip_list=None
    ):
        schema_editor = connection.SchemaEditorClass(connection)
        for field in all_local_fields(meta):
            if skip_list and field.attname in skip_list:
                continue
            if field.unique and meta.managed:
                attname = field.db_column or field.attname
                db_field_unique = table_indexes.get(attname, {}).get("unique")
                if not db_field_unique and table_constraints:
                    db_field_unique = any(
                        constraint["unique"]
                        for contraint_name, constraint in table_constraints.items()
                        if [attname] == constraint["columns"]
                    )
                if attname in table_indexes and db_field_unique:
                    continue

                index_name = schema_editor._create_index_name(table_name, [attname])

                self.add_difference(
                    "unique-missing-in-db", table_name, [attname], index_name + "_uniq"
                )
                db_type = field.db_type(connection=connection)
                if db_type.startswith("varchar"):
                    self.add_difference(
                        "index-missing-in-db",
                        table_name,
                        [attname],
                        index_name + "_like",
                        " varchar_pattern_ops",
                    )
                if db_type.startswith("text"):
                    self.add_difference(
                        "index-missing-in-db",
                        table_name,
                        [attname],
                        index_name + "_like",
                        " text_pattern_ops",
                    )

        unique_together = self.get_unique_together(meta)
        db_unique_columns = normalize_together(
            [
                v["columns"]
                for v in table_constraints.values()
                if v["unique"] and not v["index"]
            ]
        )

        for unique_columns in unique_together:
            if unique_columns in db_unique_columns:
                continue

            if skip_list and unique_columns in skip_list:
                continue

            index_name = schema_editor._create_index_name(table_name, unique_columns)

            self.add_difference(
                "unique-missing-in-db", table_name, unique_columns, index_name + "_uniq"
            )

    def find_unique_missing_in_model(
        self, meta, table_indexes, table_constraints, table_name
    ):
        fields = dict([(field.column, field) for field in all_local_fields(meta)])
        unique_together = self.get_unique_together(meta)

        for constraint_name, constraint in table_constraints.items():
            if not constraint["unique"]:
                continue
            if constraint["index"]:
                # unique indexes are handled by find_index_missing_in_model
                continue

            columns = constraint["columns"]
            if len(columns) == 1:
                field = fields.get(columns[0])
                if field is None:
                    pass
                elif field.unique:
                    continue
            else:
                if tuple(columns) in unique_together:
                    continue

            self.add_difference("unique-missing-in-model", table_name, constraint_name)

    def find_index_missing_in_db(
        self, meta, table_indexes, table_constraints, table_name
    ):
        schema_editor = connection.SchemaEditorClass(connection)
        for field in all_local_fields(meta):
            if field.db_index:
                attname = field.db_column or field.attname
                if attname not in table_indexes:
                    index_name = schema_editor._create_index_name(table_name, [attname])
                    self.add_difference(
                        "index-missing-in-db", table_name, [attname], index_name, ""
                    )
                    db_type = field.db_type(connection=connection)
                    if db_type.startswith("varchar"):
                        self.add_difference(
                            "index-missing-in-db",
                            table_name,
                            [attname],
                            index_name + "_like",
                            " varchar_pattern_ops",
                        )
                    if db_type.startswith("text"):
                        self.add_difference(
                            "index-missing-in-db",
                            table_name,
                            [attname],
                            index_name + "_like",
                            " text_pattern_ops",
                        )

        index_together = self.get_index_together(meta)
        db_index_together = normalize_together(
            [
                v["columns"]
                for v in table_constraints.values()
                if v["index"] and not v["unique"]
            ]
        )
        for columns in index_together:
            if columns in db_index_together:
                continue
            index_name = schema_editor._create_index_name(table_name, columns)
            self.add_difference(
                "index-missing-in-db", table_name, columns, index_name + "_idx", ""
            )

        for index in meta.indexes:
            if index.name not in table_constraints:
                self.add_difference(
                    "index-missing-in-db", table_name, index.fields, index.name, ""
                )

    def find_index_missing_in_model(
        self, meta, table_indexes, table_constraints, table_name
    ):
        fields = dict([(field.column, field) for field in all_local_fields(meta)])
        meta_index_names = [idx.name for idx in meta.indexes]
        index_together = self.get_index_together(meta)

        for constraint_name, constraint in table_constraints.items():
            if constraint_name in meta_index_names:
                continue
            if constraint["unique"] and not constraint["index"]:
                # unique constraints are handled by find_unique_missing_in_model
                continue

            columns = constraint["columns"]
            field = fields.get(columns[0])
            if (constraint["unique"] and constraint["index"]) or field is None:
                # unique indexes do not exist in django ? only unique constraints
                pass
            elif len(columns) == 1:
                if constraint["primary_key"] and field.primary_key:
                    continue
                if (
                    constraint["foreign_key"]
                    and isinstance(field, models.ForeignKey)
                    and field.db_constraint
                ):
                    continue
                if constraint["unique"] and field.unique:
                    continue
                if (
                    constraint["index"]
                    and constraint["type"] == "idx"
                    and constraint.get("orders")
                    and field.unique
                ):
                    # django automatically creates a _like varchar_pattern_ops
                    #  / text_pattern_ops index see https://code.djangoproject.com/ticket/12234
                    # note: mysql does not have and/or introspect and fill the 'orders'
                    #       attribute of constraint information
                    continue
                if constraint["index"] and field.db_index:
                    continue
                if constraint["check"] and field.db_check(connection=connection):
                    continue
                if getattr(field, "spatial_index", False):
                    continue
            else:
                if constraint["index"] and tuple(columns) in index_together:
                    continue

            self.add_difference("index-missing-in-model", table_name, constraint_name)

    def find_field_missing_in_model(self, fieldmap, table_description, table_name):
        for row in table_description:
            if row[0] not in fieldmap:
                self.add_difference("field-missing-in-model", table_name, row[0])

    def find_field_missing_in_db(self, fieldmap, table_description, table_name):
        db_fields = [row[0] for row in table_description]
        for field_name, field in fieldmap.items():
            if field_name not in db_fields:
                field_output = []

                if field.remote_field:
                    field_output.extend(
                        [
                            field.remote_field.model._meta.db_table,
                            field.remote_field.model._meta.get_field(
                                field.remote_field.field_name
                            ).column,
                        ]
                    )
                    op = "fkey-missing-in-db"
                else:
                    op = "field-missing-in-db"
                field_output.append(field.db_type(connection=connection))
                if self.options["include_defaults"] and field.has_default():
                    field_output.append(
                        "DEFAULT %s" % field.get_prep_value(field.get_default())
                    )
                if not field.null:
                    field_output.append("NOT NULL")
                self.add_difference(op, table_name, field_name, *field_output)
                self.new_db_fields.add((table_name, field_name))

    def find_field_type_differ(self, meta, table_description, table_name, func=None):
        db_fields = dict([(row[0], row) for row in table_description])
        for field in all_local_fields(meta):
            if field.name not in db_fields:
                continue
            description = db_fields[field.name]

            model_type = self.get_field_model_type(field)
            db_type = self.get_field_db_type(description, field, table_name)

            # use callback function if defined
            if func:
                model_type, db_type = func(field, description, model_type, db_type)

            if not self.strip_parameters(db_type) == self.strip_parameters(
                model_type
            ) and (db_type, model_type) not in {
                ("serial", "integer"),
                ("bigserial", "bigint"),
            }:
                self.add_difference(
                    "field-type-differ", table_name, field.name, model_type, db_type
                )

    def find_field_parameter_differ(
        self, meta, table_description, table_name, func=None
    ):
        db_fields = dict([(row[0], row) for row in table_description])
        for field in all_local_fields(meta):
            if field.name not in db_fields:
                continue
            description = db_fields[field.name]

            model_type = self.get_field_model_type(field)
            db_type = self.get_field_db_type(description, field, table_name)

            if not self.strip_parameters(model_type) == self.strip_parameters(db_type):
                continue

            # use callback function if defined
            if func:
                model_type, db_type = func(field, description, model_type, db_type)

            model_check = field.db_parameters(connection=connection)["check"]
            if " CHECK" in db_type:
                db_type, db_check = db_type.split(" CHECK", 1)
                db_check = db_check.strip().lstrip("(").rstrip(")")
            else:
                db_check = None

            if not model_type == db_type or not model_check == db_check:
                self.add_difference(
                    "field-parameter-differ",
                    table_name,
                    field.name,
                    model_type,
                    db_type,
                )

    def find_field_notnull_differ(self, meta, table_description, table_name):
        if not self.can_detect_notnull_differ:
            return

        for field in all_local_fields(meta):
            attname = field.db_column or field.attname
            if (table_name, attname) in self.new_db_fields:
                continue
            null = self.get_field_db_nullable(field, table_name)
            if field.null != null:
                action = field.null and "DROP" or "SET"
                self.add_difference("notnull-differ", table_name, attname, action)

    def get_constraints(self, cursor, table_name, introspection):
        return {}

    def find_differences(self):
        if self.options["all_applications"]:
            self.add_app_model_marker(None, None)
            for table in self.db_tables:
                if (
                    table not in self.django_tables
                    and table not in self.IGNORE_MISSING_TABLES
                ):
                    self.add_difference("table-missing-in-model", table)

        cur_app_label = None
        for app_model in self.app_models:
            meta = app_model._meta
            table_name = meta.db_table
            app_label = meta.app_label

            if not self.options["include_proxy_models"] and meta.proxy:
                continue

            if cur_app_label != app_label:
                # Marker indicating start of difference scan for this table_name
                self.add_app_model_marker(app_label, app_model.__name__)

            if table_name not in self.db_tables:
                # Table is missing from database
                self.add_difference("table-missing-in-db", table_name)
                continue

            if hasattr(self.introspection, "get_constraints"):
                table_constraints = self.introspection.get_constraints(
                    self.cursor, table_name
                )
            else:
                table_constraints = self.get_constraints(
                    self.cursor, table_name, self.introspection
                )

            fieldmap = dict(
                [
                    (field.db_column or field.get_attname(), field)
                    for field in all_local_fields(meta)
                ]
            )

            # add ordering field if model uses order_with_respect_to
            if meta.order_with_respect_to:
                fieldmap["_order"] = ORDERING_FIELD

            try:
                table_description = self.introspection.get_table_description(
                    self.cursor, table_name
                )
            except Exception as e:
                self.add_difference(
                    "error", "unable to introspect table: %s" % str(e).strip()
                )
                transaction.rollback()  # reset transaction
                continue

            # map table_constraints into table_indexes
            table_indexes = {}
            for contraint_name, dct in table_constraints.items():
                columns = dct["columns"]
                if len(columns) == 1:
                    table_indexes[columns[0]] = {
                        "primary_key": dct["primary_key"],
                        "unique": dct["unique"],
                        "type": dct.get("type"),
                        "contraint_name": contraint_name,
                    }

            # Fields which are defined in database but not in model
            # 1) find: 'unique-missing-in-model'
            self.find_unique_missing_in_model(
                meta, table_indexes, table_constraints, table_name
            )
            # 2) find: 'index-missing-in-model'
            self.find_index_missing_in_model(
                meta, table_indexes, table_constraints, table_name
            )
            # 3) find: 'field-missing-in-model'
            self.find_field_missing_in_model(fieldmap, table_description, table_name)

            # Fields which are defined in models but not in database
            # 4) find: 'field-missing-in-db'
            self.find_field_missing_in_db(fieldmap, table_description, table_name)
            # 5) find: 'unique-missing-in-db'
            self.find_unique_missing_in_db(
                meta, table_indexes, table_constraints, table_name
            )
            # 6) find: 'index-missing-in-db'
            self.find_index_missing_in_db(
                meta, table_indexes, table_constraints, table_name
            )

            # Fields which have a different type or parameters
            # 7) find: 'type-differs'
            self.find_field_type_differ(meta, table_description, table_name)
            # 8) find: 'type-parameter-differs'
            self.find_field_parameter_differ(meta, table_description, table_name)
            # 9) find: 'field-notnull'
            self.find_field_notnull_differ(meta, table_description, table_name)
        self.has_differences = max(
            [len(diffs) for _app_label, _model_name, diffs in self.differences]
        )

    def print_diff(self, style=no_style()):
        """Print differences to stdout"""
        if self.options["sql"]:
            self.print_diff_sql(style)
        else:
            self.print_diff_text(style)

    def print_diff_text(self, style):
        if not self.can_detect_notnull_differ:
            self.stdout.write(
                style.NOTICE(
                    "# Detecting notnull changes not implemented for this "
                    "database backend"
                )
            )
            self.stdout.write("")

        if not self.can_detect_unsigned_differ:
            self.stdout.write(
                style.NOTICE(
                    "# Detecting unsigned changes not implemented for this "
                    "database backend"
                )
            )
            self.stdout.write("")

        cur_app_label = None
        for app_label, model_name, diffs in self.differences:
            if not diffs:
                continue
            if not self.dense and app_label and cur_app_label != app_label:
                self.stdout.write(
                    "%s %s"
                    % (style.NOTICE("+ Application:"), style.SQL_TABLE(app_label))
                )
                cur_app_label = app_label
            if not self.dense and model_name:
                self.stdout.write(
                    "%s %s"
                    % (
                        style.NOTICE("|-+ Differences for model:"),
                        style.SQL_TABLE(model_name),
                    )
                )
            for diff in diffs:
                diff_type, diff_args = diff
                text = self.DIFF_TEXTS[diff_type] % dict(
                    (
                        str(i),
                        style.SQL_TABLE(
                            ", ".join(e) if isinstance(e, (list, tuple)) else e
                        ),
                    )
                    for i, e in enumerate(diff_args)
                )
                text = "'".join(
                    i % 2 == 0 and style.ERROR(e) or e
                    for i, e in enumerate(text.split("'"))
                )
                if not self.dense:
                    self.stdout.write("%s %s" % (style.NOTICE("|--+"), text))
                else:
                    if app_label:
                        self.stdout.write(
                            "%s %s %s %s %s"
                            % (
                                style.NOTICE("App"),
                                style.SQL_TABLE(app_label),
                                style.NOTICE("Model"),
                                style.SQL_TABLE(model_name),
                                text,
                            )
                        )
                    else:
                        self.stdout.write(text)

    def print_diff_sql(self, style):
        if not self.can_detect_notnull_differ:
            self.stdout.write(
                style.NOTICE(
                    "-- Detecting notnull changes not implemented for this "
                    "database backend"
                )
            )
            self.stdout.write("")

        cur_app_label = None
        qn = connection.ops.quote_name
        if not self.has_differences:
            if not self.dense:
                self.stdout.write(style.SQL_KEYWORD("-- No differences"))
        else:
            self.stdout.write(style.SQL_KEYWORD("BEGIN;"))
            for app_label, model_name, diffs in self.differences:
                if not diffs:
                    continue
                if not self.dense and cur_app_label != app_label:
                    self.stdout.write(
                        style.NOTICE("-- Application: %s" % style.SQL_TABLE(app_label))
                    )
                    cur_app_label = app_label
                if not self.dense and model_name:
                    self.stdout.write(
                        style.NOTICE("-- Model: %s" % style.SQL_TABLE(model_name))
                    )
                for diff in diffs:
                    diff_type, diff_args = diff
                    text = self.DIFF_SQL[diff_type](style, qn, diff_args)
                    if self.dense:
                        text = text.replace("\n\t", " ")
                    self.stdout.write(text)
            self.stdout.write(style.SQL_KEYWORD("COMMIT;"))


class GenericSQLDiff(SQLDiff):
    can_detect_notnull_differ = False
    can_detect_unsigned_differ = False

    def load_null(self):
        pass

    def load_unsigned(self):
        pass


class MySQLDiff(SQLDiff):
    can_detect_notnull_differ = True
    can_detect_unsigned_differ = True
    unsigned_suffix = "UNSIGNED"

    def load(self):
        super().load()
        self.auto_increment = set()
        self.load_auto_increment()

    def format_field_names(self, field_names):
        return [f.lower() for f in field_names]

    def load_null(self):
        tablespace = "public"
        for table_name in self.db_tables:
            result = self.sql_to_dict(
                """
                SELECT column_name, is_nullable
                FROM information_schema.columns
                WHERE table_schema = DATABASE()
                    AND table_name = %s""",
                [table_name],
            )
            for table_info in result:
                key = (tablespace, table_name, table_info["column_name"])
                self.null[key] = table_info["is_nullable"] == "YES"

    def load_unsigned(self):
        tablespace = "public"
        for table_name in self.db_tables:
            result = self.sql_to_dict(
                """
                SELECT column_name
                FROM information_schema.columns
                WHERE table_schema = DATABASE()
                    AND table_name = %s
                    AND column_type LIKE '%%unsigned'""",
                [table_name],
            )
            for table_info in result:
                key = (tablespace, table_name, table_info["column_name"])
                self.unsigned.add(key)

    def load_auto_increment(self):
        for table_name in self.db_tables:
            result = self.sql_to_dict(
                """
                SELECT column_name
                FROM information_schema.columns
                WHERE table_schema = DATABASE()
                   AND table_name = %s
                   AND extra = 'auto_increment'""",
                [table_name],
            )
            for table_info in result:
                key = (table_name, table_info["column_name"])
                self.auto_increment.add(key)

    # All the MySQL hacks together create something of a problem
    # Fixing one bug in MySQL creates another issue. So just keep in mind
    # that this is way unreliable for MySQL atm.
    def get_field_db_type(self, description, field=None, table_name=None):
        db_type = super().get_field_db_type(description, field, table_name)
        if not db_type:
            return
        if field:
            # MySQL isn't really sure about char's and varchar's like sqlite
            field_type = self.get_field_model_type(field)

            # Fix char/varchar inconsistencies
            if (
                self.strip_parameters(field_type) == "char"
                and self.strip_parameters(db_type) == "varchar"
            ):
                db_type = db_type.lstrip("var")

            # They like to call bools various integer types and introspection makes
            # that a integer just convert them all to bools
            if self.strip_parameters(field_type) == "bool":
                if db_type == "integer":
                    db_type = "bool"

            if (
                table_name,
                field.column,
            ) in self.auto_increment and "AUTO_INCREMENT" not in db_type:
                db_type += " AUTO_INCREMENT"
        return db_type

    def find_index_missing_in_model(
        self, meta, table_indexes, table_constraints, table_name
    ):
        fields = dict([(field.column, field) for field in all_local_fields(meta)])
        meta_index_names = [idx.name for idx in meta.indexes]
        index_together = self.get_index_together(meta)
        unique_together = self.get_unique_together(meta)

        for constraint_name, constraint in table_constraints.items():
            if constraint_name in meta_index_names:
                continue
            if constraint["unique"] and not constraint["index"]:
                # unique constraints are handled by find_unique_missing_in_model
                continue

            columns = constraint["columns"]
            field = fields.get(columns[0])

            # extra check removed from superclass here, otherwise function is the same
            if len(columns) == 1:
                if not field:
                    # both index and field are missing from the model
                    self.add_difference(
                        "index-missing-in-model", table_name, constraint_name
                    )
                    continue
                if constraint["primary_key"] and field.primary_key:
                    continue
                if (
                    constraint["foreign_key"]
                    and isinstance(field, models.ForeignKey)
                    and field.db_constraint
                ):
                    continue
                if constraint["unique"] and field.unique:
                    continue
                if (
                    constraint["index"]
                    and constraint["type"] == "idx"
                    and constraint.get("orders")
                    and field.unique
                ):
                    # django automatically creates a _like varchar_pattern_ops
                    #  / text_pattern_ops index see https://code.djangoproject.com/ticket/12234
                    # note: mysql does not have and/or introspect and fill the 'orders'
                    #       attribute of constraint information
                    continue
                if constraint["index"] and field.db_index:
                    continue
                if constraint["check"] and field.db_check(connection=connection):
                    continue
                if getattr(field, "spatial_index", False):
                    continue
            else:
                if constraint["index"] and tuple(columns) in index_together:
                    continue
                if (
                    constraint["index"]
                    and constraint["unique"]
                    and tuple(columns) in unique_together
                ):
                    continue

            self.add_difference("index-missing-in-model", table_name, constraint_name)

    def find_unique_missing_in_db(
        self, meta, table_indexes, table_constraints, table_name, skip_list=None
    ):
        schema_editor = connection.SchemaEditorClass(connection)
        for field in all_local_fields(meta):
            if skip_list and field.attname in skip_list:
                continue
            if field.unique and meta.managed:
                attname = field.db_column or field.attname
                db_field_unique = table_indexes.get(attname, {}).get("unique")
                if not db_field_unique and table_constraints:
                    db_field_unique = any(
                        constraint["unique"]
                        for contraint_name, constraint in table_constraints.items()
                        if [attname] == constraint["columns"]
                    )
                if attname in table_indexes and db_field_unique:
                    continue

                index_name = schema_editor._create_index_name(table_name, [attname])

                self.add_difference(
                    "unique-missing-in-db", table_name, [attname], index_name + "_uniq"
                )
                db_type = field.db_type(connection=connection)
                if db_type.startswith("varchar"):
                    self.add_difference(
                        "index-missing-in-db",
                        table_name,
                        [attname],
                        index_name + "_like",
                        " varchar_pattern_ops",
                    )
                if db_type.startswith("text"):
                    self.add_difference(
                        "index-missing-in-db",
                        table_name,
                        [attname],
                        index_name + "_like",
                        " text_pattern_ops",
                    )

        unique_together = self.get_unique_together(meta)

        # This comparison changed from superclass - otherwise function is the same
        db_unique_columns = normalize_together(
            [v["columns"] for v in table_constraints.values() if v["unique"]]
        )

        for unique_columns in unique_together:
            if unique_columns in db_unique_columns:
                continue

            if skip_list and unique_columns in skip_list:
                continue

            index_name = schema_editor._create_index_name(table_name, unique_columns)
            self.add_difference(
                "unique-missing-in-db", table_name, unique_columns, index_name + "_uniq"
            )


class SqliteSQLDiff(SQLDiff):
    can_detect_notnull_differ = True
    can_detect_unsigned_differ = False

    def load_null(self):
        for table_name in self.db_tables:
            # sqlite does not support tablespaces
            tablespace = "public"
            # index, column_name, column_type, nullable, default_value
            # see: https://www.sqlite.org/pragma.html#pragma_table_info
            for table_info in self.sql_to_dict(
                "PRAGMA table_info('%s');" % table_name, []
            ):
                key = (tablespace, table_name, table_info["name"])
                self.null[key] = not table_info["notnull"]

    def load_unsigned(self):
        pass

    # Unique does not seem to be implied on Sqlite for Primary_key's
    # if this is more generic among databases this might be usefull
    # to add to the superclass's find_unique_missing_in_db method
    def find_unique_missing_in_db(
        self, meta, table_indexes, table_constraints, table_name, skip_list=None
    ):
        if skip_list is None:
            skip_list = []

        unique_columns = [
            field.db_column or field.attname
            for field in all_local_fields(meta)
            if field.unique
        ]

        for constraint in table_constraints.values():
            columns = constraint["columns"]
            if len(columns) == 1:
                column = columns[0]
                if column in unique_columns and (
                    constraint["unique"] or constraint["primary_key"]
                ):
                    skip_list.append(column)

        unique_together = self.get_unique_together(meta)
        db_unique_columns = normalize_together(
            [v["columns"] for v in table_constraints.values() if v["unique"]]
        )

        for unique_columns in unique_together:
            if unique_columns in db_unique_columns:
                skip_list.append(unique_columns)

        super().find_unique_missing_in_db(
            meta, table_indexes, table_constraints, table_name, skip_list=skip_list
        )

    # Finding Indexes by using the get_indexes dictionary doesn't seem to work
    # for sqlite.
    def find_index_missing_in_db(
        self, meta, table_indexes, table_constraints, table_name
    ):
        pass

    def find_index_missing_in_model(
        self, meta, table_indexes, table_constraints, table_name
    ):
        pass

    def get_field_db_type(self, description, field=None, table_name=None):
        db_type = super().get_field_db_type(description, field, table_name)
        if not db_type:
            return None
        if field:
            field_type = self.get_field_model_type(field)
            # Fix char/varchar inconsistencies
            if (
                self.strip_parameters(field_type) == "char"
                and self.strip_parameters(db_type) == "varchar"
            ):
                db_type = db_type.lstrip("var")
        return db_type


class PostgresqlSQLDiff(SQLDiff):
    can_detect_notnull_differ = True
    can_detect_unsigned_differ = True

    DATA_TYPES_REVERSE_NAME = {
        "hstore": "django.contrib.postgres.fields.HStoreField",
        "jsonb": "django.contrib.postgres.fields.JSONField",
    }

    # Hopefully in the future we can add constraint checking and other more
    # advanced checks based on this database.
    SQL_LOAD_CONSTRAINTS = """
        SELECT nspname, relname, conname, attname, pg_get_constraintdef(pg_constraint.oid)
        FROM pg_constraint
        INNER JOIN pg_attribute ON pg_constraint.conrelid = pg_attribute.attrelid AND pg_attribute.attnum = any(pg_constraint.conkey)
        INNER JOIN pg_class ON conrelid=pg_class.oid
        INNER JOIN pg_namespace ON pg_namespace.oid=pg_class.relnamespace
        ORDER BY CASE WHEN contype='f' THEN 0 ELSE 1 END,contype,nspname,relname,conname;
    """  # noqa: E501
    SQL_LOAD_NULL = """
        SELECT nspname, relname, attname, attnotnull
        FROM pg_attribute
        INNER JOIN pg_class ON attrelid=pg_class.oid
        INNER JOIN pg_namespace ON pg_namespace.oid=pg_class.relnamespace;
    """

    SQL_FIELD_TYPE_DIFFER = lambda self, style, qn, args: "%s %s\n\t%s %s %s %s;" % (
        style.SQL_KEYWORD("ALTER TABLE"),
        style.SQL_TABLE(qn(args[0])),
        style.SQL_KEYWORD("ALTER"),
        style.SQL_FIELD(qn(args[1])),
        style.SQL_KEYWORD("TYPE"),
        style.SQL_COLTYPE(args[2]),
    )
    SQL_FIELD_PARAMETER_DIFFER = (
        lambda self, style, qn, args: "%s %s\n\t%s %s %s %s;"
        % (
            style.SQL_KEYWORD("ALTER TABLE"),
            style.SQL_TABLE(qn(args[0])),
            style.SQL_KEYWORD("ALTER"),
            style.SQL_FIELD(qn(args[1])),
            style.SQL_KEYWORD("TYPE"),
            style.SQL_COLTYPE(args[2]),
        )
    )
    SQL_NOTNULL_DIFFER = lambda self, style, qn, args: "%s %s\n\t%s %s %s %s;" % (
        style.SQL_KEYWORD("ALTER TABLE"),
        style.SQL_TABLE(qn(args[0])),
        style.SQL_KEYWORD("ALTER COLUMN"),
        style.SQL_FIELD(qn(args[1])),
        style.SQL_KEYWORD(args[2]),
        style.SQL_KEYWORD("NOT NULL"),
    )

    def load(self):
        super().load()
        self.check_constraints = {}
        self.load_constraints()

    def load_null(self):
        for dct in self.sql_to_dict(self.SQL_LOAD_NULL, []):
            key = (dct["nspname"], dct["relname"], dct["attname"])
            self.null[key] = not dct["attnotnull"]

    def load_unsigned(self):
        # PostgreSQL does not support unsigned, so no columns are
        # unsigned. Nothing to do.
        pass

    def load_constraints(self):
        for dct in self.sql_to_dict(self.SQL_LOAD_CONSTRAINTS, []):
            key = (dct["nspname"], dct["relname"], dct["attname"])
            if "CHECK" in dct["pg_get_constraintdef"]:
                self.check_constraints[key] = dct

    def get_data_type_arrayfield(self, base_field):
        return {
            "name": "django.contrib.postgres.fields.ArrayField",
            "kwargs": {
                "base_field": self.get_field_class(base_field)(),
            },
        }

    def get_data_types_reverse_override(self):
        return {
            1042: "CharField",
            1000: lambda: self.get_data_type_arrayfield(base_field="BooleanField"),
            1001: lambda: self.get_data_type_arrayfield(base_field="BinaryField"),
            1002: lambda: self.get_data_type_arrayfield(base_field="CharField"),
            1005: lambda: self.get_data_type_arrayfield(base_field="IntegerField"),
            1006: lambda: self.get_data_type_arrayfield(base_field="IntegerField"),
            1007: lambda: self.get_data_type_arrayfield(base_field="IntegerField"),
            1009: lambda: self.get_data_type_arrayfield(base_field="CharField"),
            1014: lambda: self.get_data_type_arrayfield(base_field="CharField"),
            1015: lambda: self.get_data_type_arrayfield(base_field="CharField"),
            1016: lambda: self.get_data_type_arrayfield(base_field="BigIntegerField"),
            1017: lambda: self.get_data_type_arrayfield(base_field="FloatField"),
            1021: lambda: self.get_data_type_arrayfield(base_field="FloatField"),
            1022: lambda: self.get_data_type_arrayfield(base_field="FloatField"),
            1115: lambda: self.get_data_type_arrayfield(base_field="DateTimeField"),
            1185: lambda: self.get_data_type_arrayfield(base_field="DateTimeField"),
            1231: lambda: self.get_data_type_arrayfield(base_field="DecimalField"),
            # {'name': 'django.contrib.postgres.fields.ArrayField', 'kwargs': {'base_field': 'IntegerField'}},  # noqa: E501
            1186: lambda: self.get_data_type_arrayfield(base_field="DurationField"),
            # 1186: 'django.db.models.fields.DurationField',
            3614: "django.contrib.postgres.search.SearchVectorField",
            3802: "django.contrib.postgres.fields.JSONField",
        }

    def get_constraints(self, cursor, table_name, introspection):
        """
        Find constraints for table

        Backport of django's introspection.get_constraints(...)
        """
        constraints = {}
        # Loop over the key table, collecting things as constraints
        # This will get PKs, FKs, and uniques, but not CHECK
        cursor.execute(
            """
            SELECT
                kc.constraint_name,
                kc.column_name,
                c.constraint_type,
                array(SELECT table_name::text || '.' || column_name::text FROM information_schema.constraint_column_usage WHERE constraint_name = kc.constraint_name)
            FROM information_schema.key_column_usage AS kc
            JOIN information_schema.table_constraints AS c ON
                kc.table_schema = c.table_schema AND
                kc.table_name = c.table_name AND
                kc.constraint_name = c.constraint_name
            WHERE
                kc.table_schema = %s AND
                kc.table_name = %s
        """,  # noqa: E501
            ["public", table_name],
        )
        for constraint, column, kind, used_cols in cursor.fetchall():
            # If we're the first column, make the record
            if constraint not in constraints:
                constraints[constraint] = {
                    "columns": [],
                    "primary_key": kind.lower() == "primary key",
                    "unique": kind.lower() in ["primary key", "unique"],
                    "foreign_key": tuple(used_cols[0].split(".", 1))
                    if kind.lower() == "foreign key"
                    else None,
                    "check": False,
                    "index": False,
                }
            # Record the details
            constraints[constraint]["columns"].append(column)
        # Now get CHECK constraint columns
        cursor.execute(
            """
            SELECT kc.constraint_name, kc.column_name
            FROM information_schema.constraint_column_usage AS kc
            JOIN information_schema.table_constraints AS c ON
                kc.table_schema = c.table_schema AND
                kc.table_name = c.table_name AND
                kc.constraint_name = c.constraint_name
            WHERE
                c.constraint_type = 'CHECK' AND
                kc.table_schema = %s AND
                kc.table_name = %s
        """,
            ["public", table_name],
        )
        for constraint, column in cursor.fetchall():
            # If we're the first column, make the record
            if constraint not in constraints:
                constraints[constraint] = {
                    "columns": [],
                    "primary_key": False,
                    "unique": False,
                    "foreign_key": None,
                    "check": True,
                    "index": False,
                }
            # Record the details
            constraints[constraint]["columns"].append(column)
        # Now get indexes
        cursor.execute(
            """
            SELECT
                c2.relname,
                ARRAY(
                    SELECT (SELECT attname FROM pg_catalog.pg_attribute WHERE attnum = i AND attrelid = c.oid)
                    FROM unnest(idx.indkey) i
                ),
                idx.indisunique,
                idx.indisprimary
            FROM pg_catalog.pg_class c, pg_catalog.pg_class c2,
                pg_catalog.pg_index idx
            WHERE c.oid = idx.indrelid
                AND idx.indexrelid = c2.oid
                AND c.relname = %s
        """,  # noqa: E501
            [table_name],
        )
        for index, columns, unique, primary in cursor.fetchall():
            if index not in constraints:
                constraints[index] = {
                    "columns": list(columns),
                    "primary_key": primary,
                    "unique": unique,
                    "foreign_key": None,
                    "check": False,
                    "index": True,
                }
        return constraints

    # def get_field_db_type_kwargs(
    #   self, current_kwargs, description, field=None,
    #   table_name=None, reverse_type=None,
    # ):
    #     kwargs = {}
    #     if field and 'base_field' in current_kwargs:
    #         # find
    #         attname = field.db_column or field.attname
    #         introspect_db_type = self.sql_to_dict(
    #             """SELECT attname, format_type(atttypid, atttypmod) AS type
    #                 FROM   pg_attribute
    #                 WHERE  attrelid = %s::regclass
    #                 AND    attname = %s
    #                 AND    attnum > 0
    #                 AND    NOT attisdropped
    #                 ORDER  BY attnum;
    #             """,
    #             (table_name, attname)
    #         )[0]['type']
    #         # TODO: this gives the concrete type that the database uses, why not use
    #         #       this much earlier in the process to compare to whatever django
    #         #       spits out as the database type ?
    #         max_length = re.search(
    #             "character varying\((\d+)\)\[\]", introspect_db_type
    #         )
    #         if max_length:
    #             kwargs['max_length'] = max_length[1]
    #     return kwargs

    def get_field_db_type(self, description, field=None, table_name=None):
        db_type = super().get_field_db_type(description, field, table_name)
        if not db_type:
            return
        if field:
            if db_type.endswith("[]"):
                # TODO: This is a hack for array types. Ideally we either pass the
                #       correct constraints for the type in `get_data_type_arrayfield`
                #       which instantiates the array base_field or maybe even better
                #       restructure sqldiff entirely to be based around the concrete
                #       type yielded by the code below. That gives the complete type
                #       the database uses, why not use this much earlier in the process
                #       to compare to whatever django spits out as the desired database
                #       type ?
                attname = field.db_column or field.attname
                introspect_db_type = self.sql_to_dict(
                    """SELECT attname, format_type(atttypid, atttypmod) AS type
                        FROM   pg_attribute
                        WHERE  attrelid = %s::regclass
                        AND    attname = %s
                        AND    attnum > 0
                        AND    NOT attisdropped
                        ORDER  BY attnum;
                    """,
                    (table_name, attname),
                )[0]["type"]
                if introspect_db_type.startswith("character varying"):
                    introspect_db_type = introspect_db_type.replace(
                        "character varying", "varchar"
                    )

                return introspect_db_type

            if field.primary_key and isinstance(field, AutoField):
                # TODO: Django>4.1 uses int/bigint with identity columns
                #       instead of serial/bigserial
                if db_type == "integer":
                    db_type = "serial"
                elif db_type == "bigint":
                    db_type = "bigserial"
            if table_name:
                tablespace = field.db_tablespace
                if tablespace == "":
                    tablespace = "public"
                attname = field.db_column or field.attname
                check_constraint = self.check_constraints.get(
                    (tablespace, table_name, attname), {}
                ).get("pg_get_constraintdef", None)
                if check_constraint:
                    check_constraint = check_constraint.replace("((", "(")
                    check_constraint = check_constraint.replace("))", ")")
                    check_constraint = '("'.join(
                        [
                            ")" in e
                            and '" '.join(p.strip('"') for p in e.split(" ", 1))
                            or e
                            for e in check_constraint.split("(")
                        ]
                    )
                    # TODO: might be more then one constraint in definition ?
                    db_type += " " + check_constraint
        return db_type

    def get_field_db_type_lookup(self, type_code):
        try:
            name = self.sql_to_dict(
                "SELECT typname FROM pg_type WHERE typelem=%s;", [type_code]
            )[0]["typname"]
            return self.DATA_TYPES_REVERSE_NAME.get(name.strip("_"))
        except (IndexError, KeyError):
            pass

    """
    def find_field_type_differ(self, meta, table_description, table_name):
        def callback(field, description, model_type, db_type):
            if field.primary_key and db_type=='integer':
                db_type = 'serial'
            return model_type, db_type
        super().find_field_type_differ(meta, table_description, table_name, callback)
    """


DATABASE_SQLDIFF_CLASSES = {
    "postgis": PostgresqlSQLDiff,
    "postgresql_psycopg2": PostgresqlSQLDiff,
    "postgresql": PostgresqlSQLDiff,
    "mysql": MySQLDiff,
    "sqlite3": SqliteSQLDiff,
    "oracle": GenericSQLDiff,
}


class Command(BaseCommand):
    help = """Prints the (approximated) difference between models and fields in the database for the given app name(s).

It indicates how columns in the database are different from the sql that would
be generated by Django. This command is not a database migration tool. (Though
it can certainly help) It's purpose is to show the current differences as a way
to check/debug ur models compared to the real database tables and columns."""  # noqa: E501

    output_transaction = False

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument("app_label", nargs="*")
        parser.add_argument(
            "--all-applications",
            "-a",
            action="store_true",
            default=False,
            dest="all_applications",
            help="Automaticly include all application from INSTALLED_APPS.",
        )
        parser.add_argument(
            "--not-only-existing",
            "-e",
            action="store_false",
            default=True,
            dest="only_existing",
            help=(
                "Check all tables that exist in the database, not only tables "
                "that should exist based on models."
            ),
        )
        parser.add_argument(
            "--dense-output",
            "-d",
            action="store_true",
            dest="dense_output",
            default=False,
            help=(
                "Shows the output in dense format, "
                "normally output is spreaded over multiple lines."
            ),
        )
        parser.add_argument(
            "--output_text",
            "-t",
            action="store_false",
            dest="sql",
            default=True,
            help="Outputs the differences as descriptive text instead of SQL",
        )
        parser.add_argument(
            "--include-proxy-models",
            action="store_true",
            dest="include_proxy_models",
            default=False,
            help="Include proxy models in the graph",
        )
        parser.add_argument(
            "--include-defaults",
            action="store_true",
            dest="include_defaults",
            default=False,
            help="Include default values in SQL output (beta feature)",
        )
        parser.add_argument(
            "--migrate-for-tests",
            action="store_true",
            dest="migrate_for_tests",
            default=False,
            help=argparse.SUPPRESS,
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.exit_code = 1

    @signalcommand
    def handle(self, *args, **options):
        from django.conf import settings

        app_labels = options["app_label"]
        engine = None
        if hasattr(settings, "DATABASES"):
            engine = settings.DATABASES["default"]["ENGINE"]
        else:
            engine = settings.DATABASE_ENGINE

        if engine == "dummy":
            # This must be the "dummy" database backend, which means the user
            # hasn't set DATABASE_ENGINE.
            raise CommandError(
                "Django doesn't know which syntax to use for your SQL statements, "
                "because you haven't specified the DATABASE_ENGINE setting. "
                "Edit your settings file and change DATABASE_ENGINE to something like "
                "'postgresql' or 'mysql'."
            )

        if options["all_applications"]:
            app_models = apps.get_models(include_auto_created=True)
        else:
            if not app_labels:
                raise CommandError("Enter at least one appname.")

            if not isinstance(app_labels, (list, tuple, set)):
                app_labels = [app_labels]

            app_models = []
            for app_label in app_labels:
                app_config = apps.get_app_config(app_label)
                app_models.extend(app_config.get_models(include_auto_created=True))

        if not app_models:
            raise CommandError("Unable to execute sqldiff no models founds.")

        migrate_for_tests = options["migrate_for_tests"]
        if migrate_for_tests:
            from django.core.management import call_command

            call_command("migrate", *app_labels, no_input=True, run_syncdb=True)

        if not engine:
            engine = connection.__module__.split(".")[-2]

        if "." in engine:
            engine = engine.split(".")[-1]

        cls = DATABASE_SQLDIFF_CLASSES.get(engine, GenericSQLDiff)
        sqldiff_instance = cls(
            app_models, options, stdout=self.stdout, stderr=self.stderr
        )
        sqldiff_instance.load()
        sqldiff_instance.find_differences()
        if not sqldiff_instance.has_differences:
            self.exit_code = 0
        sqldiff_instance.print_diff(self.style)

    def execute(self, *args, **options):
        try:
            super().execute(*args, **options)
        except CommandError as e:
            if options["traceback"]:
                raise

            # self.stderr is not guaranteed to be set here
            stderr = getattr(self, "stderr", None)
            if not stderr:
                stderr = OutputWrapper(sys.stderr, self.style.ERROR)
            stderr.write("%s: %s" % (e.__class__.__name__, e))
            sys.exit(2)

    def run_from_argv(self, argv):
        super().run_from_argv(argv)
        sys.exit(self.exit_code)



# -*- coding: utf-8 -*-
import os
import sys
import shutil
from typing import List

from django.core.management.base import AppCommand
from django.core.management.color import color_style

from django_extensions.management.utils import _make_writeable, signalcommand


class Command(AppCommand):
    help = "Creates a Django jobs command directory structure for the given app name "
    "in the current directory."

    requires_system_checks: List[str] = []
    # Can't import settings during this command, because they haven't
    # necessarily been created.
    can_import_settings = True

    @signalcommand
    def handle_app_config(self, app, **options):
        copy_template("jobs_template", app.path, **options)


def copy_template(template_name, copy_to, **options):
    """Copy the specified template directory to the copy_to location"""
    import django_extensions

    style = color_style()
    ERROR = getattr(style, "ERROR", lambda x: x)
    SUCCESS = getattr(style, "SUCCESS", lambda x: x)

    template_dir = os.path.join(django_extensions.__path__[0], "conf", template_name)
    verbosity = options["verbosity"]

    # walks the template structure and copies it
    for d, subdirs, files in os.walk(template_dir):
        relative_dir = d[len(template_dir) + 1 :]
        if relative_dir and not os.path.exists(os.path.join(copy_to, relative_dir)):
            os.mkdir(os.path.join(copy_to, relative_dir))
        for i, subdir in enumerate(subdirs):
            if subdir.startswith("."):
                del subdirs[i]
        for f in files:
            if f.endswith(".pyc") or f.startswith(".DS_Store"):
                continue
            path_old = os.path.join(d, f)
            path_new = os.path.join(copy_to, relative_dir, f).rstrip(".tmpl")
            if os.path.exists(path_new):
                if verbosity > 1:
                    print(ERROR("%s already exists" % path_new))
                continue
            if verbosity > 1:
                print(SUCCESS("%s" % path_new))

            with open(path_old, "r") as fp_orig:
                with open(path_new, "w") as fp_new:
                    fp_new.write(fp_orig.read())

            try:
                shutil.copymode(path_old, path_new)
                _make_writeable(path_new)
            except OSError:
                sys.stderr.write(
                    "Notice: Couldn't set permission bits on %s. You're probably using an uncommon filesystem setup. No problem.\n"  # noqa: E501
                    % path_new
                )



# -*- coding: utf-8 -*-
from django.apps import apps
from django.core.management.base import CommandError, LabelCommand
from django.utils.encoding import force_str

from django_extensions.management.utils import signalcommand


class Command(LabelCommand):
    help = "Outputs the specified model as a form definition to the shell."

    def add_arguments(self, parser):
        parser.add_argument("label", type=str, help="application name and model name")
        parser.add_argument(
            "--fields",
            "-f",
            action="append",
            dest="fields",
            default=[],
            help="Describe form with these fields only",
        )

    @signalcommand
    def handle(self, *args, **options):
        label = options["label"]
        fields = options["fields"]

        return describe_form(label, fields)


def describe_form(label, fields):
    """Return a string describing a form based on the model"""
    try:
        app_name, model_name = label.split(".")[-2:]
    except (IndexError, ValueError):
        raise CommandError("Need application and model name in the form: appname.model")
    model = apps.get_model(app_name, model_name)

    opts = model._meta
    field_list = []
    for f in opts.fields + opts.many_to_many:
        if not f.editable:
            continue
        if fields and f.name not in fields:
            continue
        formfield = f.formfield()
        if "__dict__" not in dir(formfield):
            continue
        attrs = {}
        valid_fields = [
            "required",
            "initial",
            "max_length",
            "min_length",
            "max_value",
            "min_value",
            "max_digits",
            "decimal_places",
            "choices",
            "help_text",
            "label",
        ]
        for k, v in formfield.__dict__.items():
            if k in valid_fields and v is not None:
                # ignore defaults, to minimize verbosity
                if k == "required" and v:
                    continue
                if k == "help_text" and not v:
                    continue
                if k == "widget":
                    attrs[k] = v.__class__
                elif k in ["help_text", "label"]:
                    attrs[k] = str(force_str(v).strip())
                else:
                    attrs[k] = v

        params = ", ".join(["%s=%r" % (k, v) for k, v in sorted(attrs.items())])
        field_list.append(
            "    %(field_name)s = forms.%(field_type)s(%(params)s)"
            % {
                "field_name": f.name,
                "field_type": formfield.__class__.__name__,
                "params": params,
            }
        )
    return """
from django import forms
from %(app_name)s.models import %(object_name)s

class %(object_name)sForm(forms.Form):
%(field_list)s
""" % {
        "app_name": app_name,
        "object_name": opts.object_name,
        "field_list": "\n".join(field_list),
    }



# -*- coding: utf-8 -*-
"""
The Django Admin Generator is a project which can automatically generate
(scaffold) a Django Admin for you. By doing this it will introspect your
models and automatically generate an Admin with properties like:

 - `list_display` for all local fields
 - `list_filter` for foreign keys with few items
 - `raw_id_fields` for foreign keys with a lot of items
 - `search_fields` for name and `slug` fields
 - `prepopulated_fields` for `slug` fields
 - `date_hierarchy` for `created_at`, `updated_at` or `joined_at` fields

The original source and latest version can be found here:
https://github.com/WoLpH/django-admin-generator/
"""

import re

from django.apps import apps
from django.conf import settings
from django.core.management.base import LabelCommand, CommandError
from django.db import models

from django_extensions.management.utils import signalcommand

# Configurable constants
MAX_LINE_WIDTH = getattr(settings, "MAX_LINE_WIDTH", 78)
INDENT_WIDTH = getattr(settings, "INDENT_WIDTH", 4)
LIST_FILTER_THRESHOLD = getattr(settings, "LIST_FILTER_THRESHOLD", 25)
RAW_ID_THRESHOLD = getattr(settings, "RAW_ID_THRESHOLD", 100)

LIST_FILTER = getattr(
    settings,
    "LIST_FILTER",
    (
        models.DateField,
        models.DateTimeField,
        models.ForeignKey,
        models.BooleanField,
    ),
)

SEARCH_FIELD_NAMES = getattr(
    settings,
    "SEARCH_FIELD_NAMES",
    (
        "name",
        "slug",
    ),
)

DATE_HIERARCHY_NAMES = getattr(
    settings,
    "DATE_HIERARCHY_NAMES",
    (
        "joined_at",
        "updated_at",
        "created_at",
    ),
)

PREPOPULATED_FIELD_NAMES = getattr(settings, "PREPOPULATED_FIELD_NAMES", ("slug=name",))

PRINT_IMPORTS = getattr(
    settings,
    "PRINT_IMPORTS",
    """# -*- coding: utf-8 -*-
from django.contrib import admin

from .models import %(models)s
""",
)

PRINT_ADMIN_CLASS = getattr(
    settings,
    "PRINT_ADMIN_CLASS",
    """

@admin.register(%(name)s)
class %(name)sAdmin(admin.ModelAdmin):%(class_)s
""",
)

PRINT_ADMIN_PROPERTY = getattr(
    settings,
    "PRINT_ADMIN_PROPERTY",
    """
    %(key)s = %(value)s""",
)


class UnicodeMixin:
    """
    Mixin class to handle defining the proper __str__/__unicode__
    methods in Python 2 or 3.
    """

    def __str__(self):
        return self.__unicode__()


class AdminApp(UnicodeMixin):
    def __init__(self, app_config, model_res, **options):
        self.app_config = app_config
        self.model_res = model_res
        self.options = options

    def __iter__(self):
        for model in self.app_config.get_models():
            admin_model = AdminModel(model, **self.options)

            for model_re in self.model_res:
                if model_re.search(admin_model.name):
                    break
            else:
                if self.model_res:
                    continue

            yield admin_model

    def __unicode__(self):
        return "".join(self._unicode_generator())

    def _unicode_generator(self):
        models_list = [admin_model.name for admin_model in self]
        yield PRINT_IMPORTS % dict(models=", ".join(models_list))

        admin_model_names = []
        for admin_model in self:
            yield PRINT_ADMIN_CLASS % dict(
                name=admin_model.name,
                class_=admin_model,
            )
            admin_model_names.append(admin_model.name)

    def __repr__(self):
        return "<%s[%s]>" % (
            self.__class__.__name__,
            self.app.name,
        )


class AdminModel(UnicodeMixin):
    PRINTABLE_PROPERTIES = (
        "list_display",
        "list_filter",
        "raw_id_fields",
        "search_fields",
        "prepopulated_fields",
        "date_hierarchy",
    )

    def __init__(
        self,
        model,
        raw_id_threshold=RAW_ID_THRESHOLD,
        list_filter_threshold=LIST_FILTER_THRESHOLD,
        search_field_names=SEARCH_FIELD_NAMES,
        date_hierarchy_names=DATE_HIERARCHY_NAMES,
        prepopulated_field_names=PREPOPULATED_FIELD_NAMES,
        **options,
    ):
        self.model = model
        self.list_display = []
        self.list_filter = []
        self.raw_id_fields = []
        self.search_fields = []
        self.prepopulated_fields = {}
        self.date_hierarchy = None
        self.search_field_names = search_field_names
        self.raw_id_threshold = raw_id_threshold
        self.list_filter_threshold = list_filter_threshold
        self.date_hierarchy_names = date_hierarchy_names
        self.prepopulated_field_names = prepopulated_field_names

    def __repr__(self):
        return "<%s[%s]>" % (
            self.__class__.__name__,
            self.name,
        )

    @property
    def name(self):
        return self.model.__name__

    def _process_many_to_many(self, meta):
        raw_id_threshold = self.raw_id_threshold
        for field in meta.local_many_to_many:
            if hasattr(field, "remote_field"):
                related_model = getattr(
                    field.remote_field, "related_model", field.remote_field.model
                )
            else:
                raise CommandError("Unable to process ManyToMany relation")
            related_objects = related_model.objects.all()
            if related_objects[:raw_id_threshold].count() < raw_id_threshold:
                yield field.name

    def _process_fields(self, meta):
        parent_fields = meta.parents.values()
        for field in meta.fields:
            name = self._process_field(field, parent_fields)
            if name:
                yield name

    def _process_foreign_key(self, field):
        raw_id_threshold = self.raw_id_threshold
        list_filter_threshold = self.list_filter_threshold
        max_count = max(list_filter_threshold, raw_id_threshold)
        if hasattr(field, "remote_field"):
            related_model = getattr(
                field.remote_field, "related_model", field.remote_field.model
            )
        else:
            raise CommandError("Unable to process ForeignKey relation")
        related_count = related_model.objects.all()
        related_count = related_count[:max_count].count()

        if related_count >= raw_id_threshold:
            self.raw_id_fields.append(field.name)

        elif related_count < list_filter_threshold:
            self.list_filter.append(field.name)

        else:  # pragma: no cover
            pass  # Do nothing :)

    def _process_field(self, field, parent_fields):
        if field in parent_fields:
            return

        field_name = str(field.name)
        self.list_display.append(field_name)
        if isinstance(field, LIST_FILTER):
            if isinstance(field, models.ForeignKey):
                self._process_foreign_key(field)
            else:
                self.list_filter.append(field_name)

        if field.name in self.search_field_names:
            self.search_fields.append(field_name)

        return field_name

    def __unicode__(self):
        return "".join(self._unicode_generator())

    def _yield_value(self, key, value):
        if isinstance(value, (list, set, tuple)):
            return self._yield_tuple(key, tuple(value))
        elif isinstance(value, dict):
            return self._yield_dict(key, value)
        elif isinstance(value, str):
            return self._yield_string(key, value)
        else:  # pragma: no cover
            raise TypeError("%s is not supported in %r" % (type(value), value))

    def _yield_string(self, key, value, converter=repr):
        return PRINT_ADMIN_PROPERTY % dict(
            key=key,
            value=converter(value),
        )

    def _yield_dict(self, key, value):
        row_parts = []
        row = self._yield_string(key, value)
        if len(row) > MAX_LINE_WIDTH:
            row_parts.append(self._yield_string(key, "{", str))
            for k, v in value.items():
                row_parts.append("%s%r: %r" % (2 * INDENT_WIDTH * " ", k, v))

            row_parts.append(INDENT_WIDTH * " " + "}")
            row = "\n".join(row_parts)

        return row

    def _yield_tuple(self, key, value):
        row_parts = []
        row = self._yield_string(key, value)
        if len(row) > MAX_LINE_WIDTH:
            row_parts.append(self._yield_string(key, "(", str))
            for v in value:
                row_parts.append(2 * INDENT_WIDTH * " " + repr(v) + ",")

            row_parts.append(INDENT_WIDTH * " " + ")")
            row = "\n".join(row_parts)

        return row

    def _unicode_generator(self):
        self._process()
        for key in self.PRINTABLE_PROPERTIES:
            value = getattr(self, key)
            if value:
                yield self._yield_value(key, value)

    def _process(self):
        meta = self.model._meta

        self.raw_id_fields += list(self._process_many_to_many(meta))
        field_names = list(self._process_fields(meta))

        for field_name in self.date_hierarchy_names[::-1]:
            if field_name in field_names and not self.date_hierarchy:
                self.date_hierarchy = field_name
                break

        for k in sorted(self.prepopulated_field_names):
            k, vs = k.split("=", 1)
            vs = vs.split(",")
            if k in field_names:
                incomplete = False
                for v in vs:
                    if v not in field_names:
                        incomplete = True
                        break

                if not incomplete:
                    self.prepopulated_fields[k] = vs

        self.processed = True


class Command(LabelCommand):
    help = """Generate a `admin.py` file for the given app (models)"""
    # args = "[app_name]"
    can_import_settings = True

    def add_arguments(self, parser):
        parser.add_argument("app_name")
        parser.add_argument("model_name", nargs="*")
        parser.add_argument(
            "-s",
            "--search-field",
            action="append",
            default=SEARCH_FIELD_NAMES,
            help="Fields named like this will be added to `search_fields`"
            " [default: %(default)s]",
        )
        parser.add_argument(
            "-d",
            "--date-hierarchy",
            action="append",
            default=DATE_HIERARCHY_NAMES,
            help="A field named like this will be set as `date_hierarchy`"
            " [default: %(default)s]",
        )
        parser.add_argument(
            "-p",
            "--prepopulated-fields",
            action="append",
            default=PREPOPULATED_FIELD_NAMES,
            help="These fields will be prepopulated by the other field."
            "The field names can be specified like `spam=eggA,eggB,eggC`"
            " [default: %(default)s]",
        )
        parser.add_argument(
            "-l",
            "--list-filter-threshold",
            type=int,
            default=LIST_FILTER_THRESHOLD,
            metavar="LIST_FILTER_THRESHOLD",
            help="If a foreign key has less than LIST_FILTER_THRESHOLD items "
            "it will be added to `list_filter` [default: %(default)s]",
        )
        parser.add_argument(
            "-r",
            "--raw-id-threshold",
            type=int,
            default=RAW_ID_THRESHOLD,
            metavar="RAW_ID_THRESHOLD",
            help="If a foreign key has more than RAW_ID_THRESHOLD items "
            "it will be added to `list_filter` [default: %(default)s]",
        )

    @signalcommand
    def handle(self, *args, **options):
        app_name = options["app_name"]

        try:
            app = apps.get_app_config(app_name)
        except LookupError:
            self.stderr.write("This command requires an existing app name as argument")
            self.stderr.write("Available apps:")
            app_labels = [app.label for app in apps.get_app_configs()]
            for label in sorted(app_labels):
                self.stderr.write("    %s" % label)
            return

        model_res = []
        for arg in options["model_name"]:
            model_res.append(re.compile(arg, re.IGNORECASE))

        self.stdout.write(AdminApp(app, model_res, **options).__str__())



# -*- coding: utf-8 -*-
import os
import sys
import importlib
import inspect
import traceback

from argparse import ArgumentTypeError

from django.apps import apps
from django.conf import settings
from django.core.management.base import CommandError

from django_extensions.management.email_notifications import EmailNotificationCommand
from django_extensions.management.utils import signalcommand


class DirPolicyChoices:
    NONE = "none"
    EACH = "each"
    ROOT = "root"


def check_is_directory(value):
    if value is None or not os.path.isdir(value):
        raise ArgumentTypeError("%s is not a directory!" % value)
    return value


class BadCustomDirectoryException(Exception):
    def __init__(self, value):
        self.message = (
            value + " If --dir-policy is custom than you must set correct directory in "
            "--dir option or in settings.RUNSCRIPT_CHDIR"
        )

    def __str__(self):
        return self.message


class Command(EmailNotificationCommand):
    help = "Runs a script in django context."

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.current_directory = os.getcwd()
        self.last_exit_code = 0

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument("script", nargs="+")
        parser.add_argument(
            "--fixtures",
            action="store_true",
            dest="infixtures",
            default=False,
            help="Also look in app.fixtures subdir",
        )
        parser.add_argument(
            "--noscripts",
            action="store_true",
            dest="noscripts",
            default=False,
            help="Do not look in app.scripts subdir",
        )
        parser.add_argument(
            "-s",
            "--silent",
            action="store_true",
            dest="silent",
            default=False,
            help="Run silently, do not show errors and tracebacks."
            " Also implies --continue-on-error.",
        )
        parser.add_argument(
            "-c",
            "--continue-on-error",
            action="store_true",
            dest="continue_on_error",
            default=False,
            help="Continue executing other scripts even though one has failed. "
            "It will print a traceback unless --no-traceback or --silent are given "
            "The exit code used when terminating will always be 1.",
        )
        parser.add_argument(
            "--no-traceback",
            action="store_true",
            dest="no_traceback",
            default=False,
            help="Do not show tracebacks",
        )
        parser.add_argument(
            "--script-args",
            nargs="*",
            type=str,
            help="Space-separated argument list to be passed to the scripts. Note that "
            "the same arguments will be passed to all named scripts.",
        )
        parser.add_argument(
            "--dir-policy",
            type=str,
            choices=[
                DirPolicyChoices.NONE,
                DirPolicyChoices.EACH,
                DirPolicyChoices.ROOT,
            ],
            help="Policy of selecting scripts execution directory: "
            "none - start all scripts in current directory "
            "each - start all scripts in their directories "
            "root - start all scripts in BASE_DIR directory ",
        )
        parser.add_argument(
            "--chdir",
            type=check_is_directory,
            help="If dir-policy option is set to custom, than this option determines "
            "script execution directory.",
        )

    @signalcommand
    def handle(self, *args, **options):
        NOTICE = self.style.SQL_TABLE
        NOTICE2 = self.style.SQL_FIELD
        ERROR = self.style.ERROR
        ERROR2 = self.style.NOTICE

        subdirs = []
        scripts = options["script"]

        if not options["noscripts"]:
            subdirs.append(getattr(settings, "RUNSCRIPT_SCRIPT_DIR", "scripts"))
        if options["infixtures"]:
            subdirs.append("fixtures")
        verbosity = options["verbosity"]
        show_traceback = options["traceback"]
        no_traceback = options["no_traceback"]
        continue_on_error = options["continue_on_error"]
        if no_traceback:
            show_traceback = False
        else:
            show_traceback = True
        silent = options["silent"]
        if silent:
            verbosity = 0
            continue_on_error = True
        email_notifications = options["email_notifications"]

        if len(subdirs) < 1:
            print(NOTICE("No subdirs to run left."))
            return

        if len(scripts) < 1:
            print(ERROR("Script name required."))
            return

        def get_directory_from_chdir():
            directory = options["chdir"] or getattr(settings, "RUNSCRIPT_CHDIR", None)
            try:
                check_is_directory(directory)
            except ArgumentTypeError as e:
                raise BadCustomDirectoryException(str(e))
            return directory

        def get_directory_basing_on_policy(script_module):
            policy = options["dir_policy"] or getattr(
                settings, "RUNSCRIPT_CHDIR_POLICY", DirPolicyChoices.NONE
            )
            if policy == DirPolicyChoices.ROOT:
                return settings.BASE_DIR
            elif policy == DirPolicyChoices.EACH:
                return os.path.dirname(inspect.getfile(script_module))
            else:
                return self.current_directory

        def set_directory(script_module):
            if options["chdir"]:
                directory = get_directory_from_chdir()
            elif options["dir_policy"]:
                directory = get_directory_basing_on_policy(script_module)
            elif getattr(settings, "RUNSCRIPT_CHDIR", None):
                directory = get_directory_from_chdir()
            else:
                directory = get_directory_basing_on_policy(script_module)
            os.chdir(os.path.abspath(directory))

        def run_script(mod, *script_args):
            exit_code = None
            try:
                set_directory(mod)
                exit_code = mod.run(*script_args)
                if isinstance(exit_code, bool):
                    # convert boolean True to exit-code 0 and False to exit-code 1
                    exit_code = 1 if exit_code else 0
                if isinstance(exit_code, int):
                    if exit_code != 0:
                        try:
                            raise CommandError(
                                "'%s' failed with exit code %s"
                                % (mod.__name__, exit_code),
                                returncode=exit_code,
                            )
                        except TypeError:
                            raise CommandError(
                                "'%s' failed with exit code %s"
                                % (mod.__name__, exit_code)
                            )
                if email_notifications:
                    self.send_email_notification(notification_id=mod.__name__)
            except Exception as e:
                if isinstance(e, CommandError) and hasattr(e, "returncode"):
                    exit_code = e.returncode
                self.last_exit_code = exit_code if isinstance(exit_code, int) else 1
                if silent:
                    return
                if verbosity > 0:
                    print(ERROR("Exception while running run() in '%s'" % mod.__name__))
                if continue_on_error:
                    if show_traceback:
                        traceback.print_exc()
                    return
                if email_notifications:
                    self.send_email_notification(
                        notification_id=mod.__name__, include_traceback=True
                    )

                if no_traceback:
                    raise CommandError(repr(e))

                raise

        def my_import(parent_package, module_name):
            full_module_path = "%s.%s" % (parent_package, module_name)
            if verbosity > 1:
                print(NOTICE("Check for %s" % full_module_path))
            # Try importing the parent package first
            try:
                importlib.import_module(parent_package)
            except ImportError as e:
                if str(e).startswith("No module named"):
                    # No need to proceed if the parent package doesn't exist
                    return False

            try:
                t = importlib.import_module(full_module_path)
            except ImportError as e:
                # The parent package exists, but the module doesn't
                try:
                    if importlib.util.find_spec(full_module_path) is None:
                        return False
                except Exception:
                    module_file = (
                        os.path.join(settings.BASE_DIR, *full_module_path.split("."))
                        + ".py"
                    )
                    if not os.path.isfile(module_file):
                        return False

                if silent:
                    return False
                if show_traceback:
                    traceback.print_exc()
                if verbosity > 0:
                    print(
                        ERROR("Cannot import module '%s': %s." % (full_module_path, e))
                    )

                return False

            if hasattr(t, "run"):
                if verbosity > 1:
                    print(NOTICE2("Found script '%s' ..." % full_module_path))
                return t
            else:
                if verbosity > 1:
                    print(
                        ERROR2(
                            "Found script '%s' but no run() function found."
                            % full_module_path
                        )
                    )

        def find_modules_for_script(script):
            """Find script module which contains 'run' attribute"""
            modules = []
            # first look in apps
            for app in apps.get_app_configs():
                for subdir in subdirs:
                    mod = my_import("%s.%s" % (app.name, subdir), script)
                    if mod:
                        modules.append(mod)
            # try direct import
            if script.find(".") != -1:
                parent, mod_name = script.rsplit(".", 1)
                mod = my_import(parent, mod_name)
                if mod:
                    modules.append(mod)
            else:
                # try app.DIR.script import
                for subdir in subdirs:
                    mod = my_import(subdir, script)
                    if mod:
                        modules.append(mod)

            return modules

        if options["script_args"]:
            script_args = options["script_args"]
        else:
            script_args = []

        # first pass to check if all scripts can be found
        script_to_run = []
        for script in scripts:
            script_modules = find_modules_for_script(script)
            if not script_modules:
                self.last_exit_code = 1
                if verbosity > 0 and not silent:
                    print(ERROR("No (valid) module for script '%s' found" % script))
                continue
            script_to_run.extend(script_modules)

        if self.last_exit_code:
            if verbosity < 2 and not silent:
                print(
                    ERROR("Try running with a higher verbosity level like: -v2 or -v3")
                )
            if not continue_on_error:
                script_to_run = []

        for script_mod in script_to_run:
            if verbosity > 1:
                print(NOTICE2("Running script '%s' ..." % script_mod.__name__))
            run_script(script_mod, *script_args)

        if self.last_exit_code != 0:
            if silent:
                if hasattr(self, "running_tests"):
                    return
                sys.exit(self.last_exit_code)

            try:
                raise CommandError(
                    "An error has occurred running scripts. See errors above.",
                    returncode=self.last_exit_code,
                )
            except TypeError:
                # Django < 3.1 fallback
                if self.last_exit_code == 1:
                    # if exit_code is 1 we can still raise CommandError without
                    # returncode argument
                    raise CommandError(
                        "An error has occurred running scripts. See errors above."
                    )
                print(ERROR("An error has occurred running scripts. See errors above."))
                if hasattr(self, "running_tests"):
                    return
                sys.exit(self.last_exit_code)



# -*- coding: utf-8 -*-
from django.apps import apps as django_apps
from django.contrib.auth.management import create_permissions
from django.contrib.auth.management import _get_all_permissions  # type: ignore
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.core.management.base import BaseCommand

from django_extensions.management.utils import signalcommand


class Command(BaseCommand):
    help = (
        "reloads permissions for specified apps, or all apps if no args are specified"
    )

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--apps",
            dest="apps",
            help="Reload permissions only for apps (comma separated)",
        )
        parser.add_argument(
            "--create-only",
            action="store_true",
            default=False,
            help="Only create missing permissions",
        )
        parser.add_argument(
            "--update-only",
            action="store_true",
            default=False,
            help="Only update permissions",
        )

    @signalcommand
    def handle(self, *args, **options):
        if options["apps"]:
            app_names = options["apps"].split(",")
            apps = [django_apps.get_app_config(x) for x in app_names]
        else:
            apps = django_apps.get_app_configs()

        if options["create_only"]:
            do_create, do_update = True, False
        elif options["update_only"]:
            do_create, do_update = False, True
        else:
            do_create, do_update = True, True

        for app in apps:
            if do_create:
                # create permissions if they do not exist
                create_permissions(app, options["verbosity"])

            if do_update:
                # update permission name's if changed
                for model in app.get_models():
                    content_type = ContentType.objects.get_for_model(model)
                    for codename, name in _get_all_permissions(model._meta):
                        try:
                            permission = Permission.objects.get(
                                codename=codename, content_type=content_type
                            )
                        except Permission.DoesNotExist:
                            continue
                        if permission.name != name:
                            old_str = str(permission)
                            permission.name = name
                            if options["verbosity"] >= 2:
                                self.stdout.write(
                                    self.style.SUCCESS(
                                        "Update permission '%s' to '%s'"
                                        % (old_str, permission)
                                    )
                                )
                            permission.save()



# -*- coding: utf-8 -*-
import os
import inspect
import re

from django.core.management.base import BaseCommand, CommandError
from django.db import DEFAULT_DB_ALIAS, connections
from django.db.migrations.loader import AmbiguityError, MigrationLoader

REPLACES_REGEX = re.compile(r"\s+replaces\s*=\s*\[[^\]]+\]\s*")
PYC = ".pyc"


def py_from_pyc(pyc_fn):
    return pyc_fn[: -len(PYC)] + ".py"


class Command(BaseCommand):
    help = (
        "Deletes left over migrations that have been replaced by a "
        "squashed migration and converts squashed migration into a normal "
        "migration. Modifies your source tree! Use with care!"
    )

    def add_arguments(self, parser):
        parser.add_argument(
            "app_label",
            help="App label of the application to delete replaced migrations from.",
        )
        parser.add_argument(
            "squashed_migration_name",
            default=None,
            nargs="?",
            help="The squashed migration to replace. "
            "If not specified defaults to the first found.",
        )
        parser.add_argument(
            "--noinput",
            "--no-input",
            action="store_false",
            dest="interactive",
            default=True,
            help="Tells Django to NOT prompt the user for input of any kind.",
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            default=False,
            help="Do not actually delete or change any files",
        )
        parser.add_argument(
            "--database",
            default=DEFAULT_DB_ALIAS,
            help=(
                "Nominates a database to run command for. "
                'Defaults to the "%s" database.'
            )
            % DEFAULT_DB_ALIAS,
        )

    def handle(self, **options):
        self.verbosity = options["verbosity"]
        self.interactive = options["interactive"]
        self.dry_run = options["dry_run"]
        app_label = options["app_label"]
        squashed_migration_name = options["squashed_migration_name"]
        database = options["database"]

        # Load the current graph state
        # check the app and migration they asked for exists
        loader = MigrationLoader(connections[database])
        if app_label not in loader.migrated_apps:
            raise CommandError(
                "App '%s' does not have migrations (so delete_squashed_migrations on "
                "it makes no sense)" % app_label
            )

        squashed_migration = None
        if squashed_migration_name:
            squashed_migration = self.find_migration(
                loader, app_label, squashed_migration_name
            )
            if not squashed_migration.replaces:
                raise CommandError(
                    "The migration %s %s is not a squashed migration."
                    % (squashed_migration.app_label, squashed_migration.name)
                )
        else:
            leaf_nodes = loader.graph.leaf_nodes(app=app_label)
            migration = loader.get_migration(*leaf_nodes[0])
            previous_migrations = [
                loader.get_migration(al, mn)
                for al, mn in loader.graph.forwards_plan(
                    (migration.app_label, migration.name)
                )
                if al == migration.app_label
            ]
            migrations = previous_migrations + [migration]
            for migration in migrations:
                if migration.replaces:
                    squashed_migration = migration
                    break

            if not squashed_migration:
                raise CommandError(
                    "Cannot find a squashed migration in app '%s'." % (app_label)
                )

        files_to_delete = []
        for al, mn in squashed_migration.replaces:
            try:
                migration = loader.disk_migrations[al, mn]
            except KeyError:
                if self.verbosity > 0:
                    self.stderr.write(
                        "Couldn't find migration file for %s %s\n" % (al, mn)
                    )
            else:
                pyc_file = inspect.getfile(migration.__class__)
                files_to_delete.append(pyc_file)
                if pyc_file.endswith(PYC):
                    py_file = py_from_pyc(pyc_file)
                    files_to_delete.append(py_file)

        # Tell them what we're doing and optionally ask if we should proceed
        if self.verbosity > 0 or self.interactive:
            self.stdout.write(
                self.style.MIGRATE_HEADING("Will delete the following files:")
            )
            for fn in files_to_delete:
                self.stdout.write(" - %s" % fn)

            if not self.confirm():
                return

        for fn in files_to_delete:
            try:
                if not self.dry_run:
                    os.remove(fn)
            except OSError:
                if self.verbosity > 0:
                    self.stderr.write("Couldn't delete %s\n" % (fn,))

        # Try and delete replaces only if it's all on one line
        squashed_migration_fn = inspect.getfile(squashed_migration.__class__)
        if squashed_migration_fn.endswith(PYC):
            squashed_migration_fn = py_from_pyc(squashed_migration_fn)
        with open(squashed_migration_fn) as fp:
            squashed_migration_lines = list(fp)

        delete_lines = []
        for i, line in enumerate(squashed_migration_lines):
            if REPLACES_REGEX.match(line):
                delete_lines.append(i)
                if i > 0 and squashed_migration_lines[i - 1].strip() == "":
                    delete_lines.insert(0, i - 1)
                break
        if not delete_lines:
            raise CommandError(
                (
                    "Couldn't find 'replaces =' line in file %s. "
                    "Please finish cleaning up manually."
                )
                % (squashed_migration_fn,)
            )

        if self.verbosity > 0 or self.interactive:
            self.stdout.write(
                self.style.MIGRATE_HEADING(
                    "Will delete line %s%s from file %s"
                    % (
                        delete_lines[0],
                        " and " + str(delete_lines[1]) if len(delete_lines) > 1 else "",
                        squashed_migration_fn,
                    )
                )
            )

            if not self.confirm():
                return

        for line_num in sorted(delete_lines, reverse=True):
            del squashed_migration_lines[line_num]

        with open(squashed_migration_fn, "w") as fp:
            if not self.dry_run:
                fp.write("".join(squashed_migration_lines))

    def confirm(self):
        if self.interactive:
            answer = None
            while not answer or answer not in "yn":
                answer = input("Do you wish to proceed? [yN] ")
                if not answer:
                    answer = "n"
                    break
                else:
                    answer = answer[0].lower()
            return answer == "y"
        return True

    def find_migration(self, loader, app_label, name):
        try:
            return loader.get_migration_by_prefix(app_label, name)
        except AmbiguityError:
            raise CommandError(
                "More than one migration matches '%s' in app '%s'. Please be "
                "more specific." % (name, app_label)
            )
        except KeyError:
            raise CommandError(
                "Cannot find a migration matching '%s' from app '%s'."
                % (name, app_label)
            )



# -*- coding: utf-8 -*-
from collections.abc import Container
import ipaddress
import itertools


class InternalIPS(Container):
    """
    InternalIPS allows to specify CIDRs for INTERNAL_IPS.

    It takes an iterable of ip addresses or ranges.

    Inspiration taken from netaddr.IPSet, please use it if you can since
    it support more advanced features like optimizing ranges and lookups.
    """

    __slots__ = ["_cidrs"]

    def __init__(self, iterable, sort_by_size=False):
        """
        Constructor.

        :param iterable: (optional) an iterable containing IP addresses and
            subnets.

        :param sort_by_size: sorts internal list according to size of ip
            ranges, largest first.
        """
        self._cidrs = []
        for address in iterable:
            self._cidrs.append(ipaddress.ip_network(address))

        if sort_by_size:
            self._cidrs = sorted(self._cidrs)

    def __contains__(self, address):
        """
        :param ip: An IP address or subnet.

        :return: ``True`` if IP address or subnet is a member of this InternalIPS set.
        """
        address = ipaddress.ip_address(address)
        for cidr in self._cidrs:
            if address in cidr:
                return True
        return False

    def __hash__(self):
        """
        Raises ``TypeError`` if this method is called.
        """
        raise TypeError("InternalIPS containers are unhashable!")

    def __len__(self):
        """
        :return: the cardinality of this InternalIPS set.
        """
        return sum(cidr.num_addresses for cidr in self._cidrs)

    def __iter__(self):
        """
        :return: an iterator over the IP addresses within this IP set.
        """
        return itertools.chain(*self._cidrs)

    def iter_cidrs(self):
        """
        :return: an iterator over individual IP subnets within this IP set.
        """
        return sorted(self._cidrs)



# -*- coding: utf-8 -*-
import unicodedata
import binascii

from django.core.exceptions import ValidationError
from django.utils.deconstruct import deconstructible
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _


@deconstructible
class NoControlCharactersValidator:
    message = _("Control Characters like new lines or tabs are not allowed.")
    code = "no_control_characters"
    whitelist = None

    def __init__(self, message=None, code=None, whitelist=None):
        if message:
            self.message = message
        if code:
            self.code = code
        if whitelist:
            self.whitelist = whitelist

    def __call__(self, value):
        value = force_str(value)
        whitelist = self.whitelist
        category = unicodedata.category
        for character in value:
            if whitelist and character in whitelist:
                continue
            if category(character)[0] == "C":
                params = {"value": value, "whitelist": whitelist}
                raise ValidationError(self.message, code=self.code, params=params)

    def __eq__(self, other):
        return (
            isinstance(other, NoControlCharactersValidator)
            and (self.whitelist == other.whitelist)
            and (self.message == other.message)
            and (self.code == other.code)
        )


@deconstructible
class NoWhitespaceValidator:
    message = _("Leading and Trailing whitespaces are not allowed.")
    code = "no_whitespace"

    def __init__(self, message=None, code=None, whitelist=None):
        if message:
            self.message = message
        if code:
            self.code = code

    def __call__(self, value):
        value = force_str(value)
        if value != value.strip():
            params = {"value": value}
            raise ValidationError(self.message, code=self.code, params=params)

    def __eq__(self, other):
        return (
            isinstance(other, NoWhitespaceValidator)
            and (self.message == other.message)
            and (self.code == other.code)
        )


@deconstructible
class HexValidator:
    messages = {
        "invalid": _("Only a hex string is allowed."),
        "length": _("Invalid length. Must be %(length)d characters."),
        "min_length": _("Ensure that there are more than %(min)s characters."),
        "max_length": _("Ensure that there are no more than %(max)s characters."),
    }
    code = "hex_only"

    def __init__(
        self, length=None, min_length=None, max_length=None, message=None, code=None
    ):
        self.length = length
        self.min_length = min_length
        self.max_length = max_length
        if message:
            self.message = message
        else:
            self.message = self.messages["invalid"]
        if code:
            self.code = code

    def __call__(self, value):
        value = force_str(value)
        if self.length and len(value) != self.length:
            raise ValidationError(
                self.messages["length"],
                code="hex_only_length",
                params={"length": self.length},
            )
        if self.min_length and len(value) < self.min_length:
            raise ValidationError(
                self.messages["min_length"],
                code="hex_only_min_length",
                params={"min": self.min_length},
            )
        if self.max_length and len(value) > self.max_length:
            raise ValidationError(
                self.messages["max_length"],
                code="hex_only_max_length",
                params={"max": self.max_length},
            )

        try:
            binascii.unhexlify(value)
        except (TypeError, binascii.Error):
            raise ValidationError(self.messages["invalid"], code="hex_only")

    def __eq__(self, other):
        return (
            isinstance(other, HexValidator)
            and (self.message == other.message)
            and (self.code == other.code)
        )



# -*- coding: utf-8 -*-
import inspect
import os
import re


from django.apps import apps
from django.core.management import color
from django.core.management import BaseCommand
from django.utils import termcolors
from django.utils.encoding import smart_str

from django_extensions.compat import load_tag_library
from django_extensions.management.color import _dummy_style_func
from django_extensions.management.utils import signalcommand


def no_style():
    style = color.no_style()
    for role in ("FILTER", "MODULE_NAME", "TAG", "TAGLIB"):
        setattr(style, role, _dummy_style_func)
    return style


def color_style():
    style = color.color_style()
    style.FILTER = termcolors.make_style(fg="yellow", opts=("bold",))
    style.MODULE_NAME = termcolors.make_style(fg="green", opts=("bold",))
    style.TAG = termcolors.make_style(fg="red", opts=("bold",))
    style.TAGLIB = termcolors.make_style(fg="blue", opts=("bold",))
    return style


def format_block(block, nlspaces=0):
    """
    Format the given block of text, trimming leading/trailing
    empty lines and any leading whitespace that is common to all lines.
    The purpose is to let us list a code block as a multiline,
    triple-quoted Python string, taking care of
    indentation concerns.
    https://code.activestate.com/recipes/145672/
    """
    # separate block into lines
    lines = smart_str(block).split("\n")

    # remove leading/trailing empty lines
    while lines and not lines[0]:
        del lines[0]
    while lines and not lines[-1]:
        del lines[-1]

    # look at first line to see how much indentation to trim
    ws = re.match(r"\s*", lines[0]).group(0)
    if ws:
        lines = map(lambda x: x.replace(ws, "", 1), lines)

    # remove leading/trailing blank lines (after leading ws removal)
    # we do this again in case there were pure-whitespace lines
    while lines and not lines[0]:
        del lines[0]
    while lines and not lines[-1]:
        del lines[-1]

    # account for user-specified leading spaces
    flines = ["%s%s" % (" " * nlspaces, line) for line in lines]

    return "\n".join(flines) + "\n"


class Command(BaseCommand):
    help = "Displays template tags and filters available in the current project."
    results = ""

    def add_result(self, s, depth=0):
        self.results += "%s\n" % s.rjust(depth * 4 + len(s))

    @signalcommand
    def handle(self, *args, **options):
        if options["no_color"]:
            style = no_style()
        else:
            style = color_style()

        for app_config in apps.get_app_configs():
            app = app_config.name
            try:
                templatetag_mod = __import__(app + ".templatetags", {}, {}, [""])
            except ImportError:
                continue
            mod_path = inspect.getabsfile(templatetag_mod)
            mod_files = os.listdir(os.path.dirname(mod_path))
            tag_files = [
                i.rstrip(".py") for i in mod_files if i.endswith(".py") and i[0] != "_"
            ]
            app_labeled = False
            for taglib in tag_files:
                lib = load_tag_library(taglib)
                if lib is None:
                    continue

                if not app_labeled:
                    self.add_result("App: %s" % style.MODULE_NAME(app))
                    app_labeled = True
                self.add_result("load: %s" % style.TAGLIB(taglib), 1)
                libstuff = [
                    (lib.tags, "Tag:", style.TAG),
                    (lib.filters, "Filter:", style.FILTER),
                ]
                for items, label, style_func in libstuff:
                    for item in items:
                        self.add_result("%s %s" % (label, style_func(item)), 2)
                        doc = inspect.getdoc(items[item])
                        if doc:
                            self.add_result(format_block(doc, 12))
        return self.results



# -*- coding: utf-8 -*-
"""
Django Extensions additional model fields

Some fields might require additional dependencies to be installed.
"""

import re
import string

try:
    import uuid

    HAS_UUID = True
except ImportError:
    HAS_UUID = False

try:
    import shortuuid

    HAS_SHORT_UUID = True
except ImportError:
    HAS_SHORT_UUID = False

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.models import DateTimeField, CharField, SlugField, Q, UniqueConstraint
from django.db.models.constants import LOOKUP_SEP
from django.template.defaultfilters import slugify
from django.utils.crypto import get_random_string
from django.utils.encoding import force_str


MAX_UNIQUE_QUERY_ATTEMPTS = getattr(
    settings, "EXTENSIONS_MAX_UNIQUE_QUERY_ATTEMPTS", 100
)


class UniqueFieldMixin:
    def check_is_bool(self, attrname):
        if not isinstance(getattr(self, attrname), bool):
            raise ValueError("'{}' argument must be True or False".format(attrname))

    @staticmethod
    def _get_fields(model_cls):
        return [
            (f, f.model if f.model != model_cls else None)
            for f in model_cls._meta.get_fields()
            if not f.is_relation or f.one_to_one or (f.many_to_one and f.related_model)
        ]

    def get_queryset(self, model_cls, slug_field):
        for field, model in self._get_fields(model_cls):
            if model and field == slug_field:
                return model._default_manager.all()
        return model_cls._default_manager.all()

    def find_unique(self, model_instance, field, iterator, *args):
        # exclude the current model instance from the queryset used in finding
        # next valid hash
        queryset = self.get_queryset(model_instance.__class__, field)
        if model_instance.pk:
            queryset = queryset.exclude(pk=model_instance.pk)

        # form a kwarg dict used to implement any unique_together constraints
        kwargs = {}
        for params in model_instance._meta.unique_together:
            if self.attname in params:
                for param in params:
                    kwargs[param] = getattr(model_instance, param, None)

        # for support django 2.2+
        query = Q()
        constraints = getattr(model_instance._meta, "constraints", None)
        if constraints:
            unique_constraints = filter(
                lambda c: isinstance(c, UniqueConstraint), constraints
            )
            for unique_constraint in unique_constraints:
                if self.attname in unique_constraint.fields:
                    condition = {
                        field: getattr(model_instance, field, None)
                        for field in unique_constraint.fields
                        if field != self.attname
                    }
                    query &= Q(**condition)

        new = next(iterator)
        kwargs[self.attname] = new
        while not new or queryset.filter(query, **kwargs):
            new = next(iterator)
            kwargs[self.attname] = new
        setattr(model_instance, self.attname, new)
        return new


class AutoSlugField(UniqueFieldMixin, SlugField):
    """
    AutoSlugField

    By default, sets editable=False, blank=True.

    Required arguments:

    populate_from
        Specifies which field, list of fields, or model method
        the slug will be populated from.

        populate_from can traverse a ForeignKey relationship
        by using Django ORM syntax:
            populate_from = 'related_model__field'

    Optional arguments:

    separator
        Defines the used separator (default: '-')

    overwrite
        If set to True, overwrites the slug on every save (default: False)

    slugify_function
        Defines the function which will be used to "slugify" a content
        (default: :py:func:`~django.template.defaultfilters.slugify` )

    It is possible to provide custom "slugify" function with
    the ``slugify_function`` function in a model class.

    ``slugify_function`` function in a model class takes priority over
    ``slugify_function`` given as an argument to :py:class:`~AutoSlugField`.

    Example

    .. code-block:: python

        # models.py

        from django.db import models

        from django_extensions.db.fields import AutoSlugField


        class MyModel(models.Model):
            def slugify_function(self, content):
                return content.replace('_', '-').lower()

            title = models.CharField(max_length=42)
            slug = AutoSlugField(populate_from='title')

    Inspired by SmileyChris' Unique Slugify snippet:
    https://www.djangosnippets.org/snippets/690/
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        kwargs.setdefault("editable", False)

        populate_from = kwargs.pop("populate_from", None)
        if populate_from is None:
            raise ValueError("missing 'populate_from' argument")
        else:
            self._populate_from = populate_from

        if not callable(populate_from):
            if not isinstance(populate_from, (list, tuple)):
                populate_from = (populate_from,)

            if not all(isinstance(e, str) for e in populate_from):
                raise TypeError(
                    "'populate_from' must be str or list[str] or tuple[str], found `%s`"
                    % populate_from
                )

        self.slugify_function = kwargs.pop("slugify_function", slugify)
        self.separator = kwargs.pop("separator", "-")
        self.overwrite = kwargs.pop("overwrite", False)
        self.check_is_bool("overwrite")
        self.overwrite_on_add = kwargs.pop("overwrite_on_add", True)
        self.check_is_bool("overwrite_on_add")
        self.allow_duplicates = kwargs.pop("allow_duplicates", False)
        self.check_is_bool("allow_duplicates")
        self.max_unique_query_attempts = kwargs.pop(
            "max_unique_query_attempts", MAX_UNIQUE_QUERY_ATTEMPTS
        )
        super().__init__(*args, **kwargs)

    def _slug_strip(self, value):
        """
        Clean up a slug by removing slug separator characters that occur at
        the beginning or end of a slug.

        If an alternate separator is used, it will also replace any instances
        of the default '-' separator with the new separator.
        """
        re_sep = "(?:-|%s)" % re.escape(self.separator)
        value = re.sub("%s+" % re_sep, self.separator, value)
        return re.sub(r"^%s+|%s+$" % (re_sep, re_sep), "", value)

    @staticmethod
    def slugify_func(content, slugify_function):
        if content:
            return slugify_function(content)
        return ""

    def slug_generator(self, original_slug, start):
        yield original_slug
        for i in range(start, self.max_unique_query_attempts):
            slug = original_slug
            end = "%s%s" % (self.separator, i)
            end_len = len(end)
            if self.slug_len and len(slug) + end_len > self.slug_len:
                slug = slug[: self.slug_len - end_len]
                slug = self._slug_strip(slug)
            slug = "%s%s" % (slug, end)
            yield slug
        raise RuntimeError(
            "max slug attempts for %s exceeded (%s)"
            % (original_slug, self.max_unique_query_attempts)
        )

    def create_slug(self, model_instance, add):
        slug = getattr(model_instance, self.attname)
        use_existing_slug = False
        if slug and not self.overwrite:
            # Existing slug and not configured to overwrite - Short-circuit
            # here to prevent slug generation when not required.
            use_existing_slug = True

        if self.overwrite_on_add and add:
            use_existing_slug = False

        if use_existing_slug:
            return slug

        # get fields to populate from and slug field to set
        populate_from = self._populate_from
        if not isinstance(populate_from, (list, tuple)):
            populate_from = (populate_from,)

        slug_field = model_instance._meta.get_field(self.attname)
        slugify_function = getattr(
            model_instance, "slugify_function", self.slugify_function
        )

        # slugify the original field content and set next step to 2
        slug_for_field = lambda lookup_value: self.slugify_func(
            self.get_slug_fields(model_instance, lookup_value),
            slugify_function=slugify_function,
        )
        slug = self.separator.join(map(slug_for_field, populate_from))
        start = 2

        # strip slug depending on max_length attribute of the slug field
        # and clean-up
        self.slug_len = slug_field.max_length
        if self.slug_len:
            slug = slug[: self.slug_len]
        slug = self._slug_strip(slug)
        original_slug = slug

        if self.allow_duplicates:
            setattr(model_instance, self.attname, slug)
            return slug

        return self.find_unique(
            model_instance, slug_field, self.slug_generator(original_slug, start)
        )

    def get_slug_fields(self, model_instance, lookup_value):
        if callable(lookup_value):
            # A function has been provided
            return "%s" % lookup_value(model_instance)

        lookup_value_path = lookup_value.split(LOOKUP_SEP)
        attr = model_instance
        for elem in lookup_value_path:
            try:
                attr = getattr(attr, elem)
            except AttributeError:
                raise AttributeError(
                    "value {} in AutoSlugField's 'populate_from' argument {} returned an error - {} has no attribute {}".format(  # noqa: E501
                        elem, lookup_value, attr, elem
                    )
                )
        if callable(attr):
            return "%s" % attr()

        return attr

    def pre_save(self, model_instance, add):
        value = force_str(self.create_slug(model_instance, add))
        return value

    def get_internal_type(self):
        return "SlugField"

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["populate_from"] = self._populate_from
        if not self.separator == "-":
            kwargs["separator"] = self.separator
        if self.overwrite is not False:
            kwargs["overwrite"] = True
        if self.allow_duplicates is not False:
            kwargs["allow_duplicates"] = True
        return name, path, args, kwargs


class RandomCharField(UniqueFieldMixin, CharField):
    """
    RandomCharField

    By default, sets editable=False, blank=True, unique=False.

    Required arguments:

    length
        Specifies the length of the field

    Optional arguments:

    unique
        If set to True, duplicate entries are not allowed (default: False)

    lowercase
        If set to True, lowercase the alpha characters (default: False)

    uppercase
        If set to True, uppercase the alpha characters (default: False)

    include_alpha
        If set to True, include alpha characters (default: True)

    include_digits
        If set to True, include digit characters (default: True)

    include_punctuation
        If set to True, include punctuation characters (default: False)

    keep_default
        If set to True, keeps the default initialization value (default: False)
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("blank", True)
        kwargs.setdefault("editable", False)

        self.length = kwargs.pop("length", None)
        if self.length is None:
            raise ValueError("missing 'length' argument")
        kwargs["max_length"] = self.length

        self.lowercase = kwargs.pop("lowercase", False)
        self.check_is_bool("lowercase")
        self.uppercase = kwargs.pop("uppercase", False)
        self.check_is_bool("uppercase")
        if self.uppercase and self.lowercase:
            raise ValueError(
                "the 'lowercase' and 'uppercase' arguments are mutually exclusive"
            )
        self.include_digits = kwargs.pop("include_digits", True)
        self.check_is_bool("include_digits")
        self.include_alpha = kwargs.pop("include_alpha", True)
        self.check_is_bool("include_alpha")
        self.include_punctuation = kwargs.pop("include_punctuation", False)
        self.keep_default = kwargs.pop("keep_default", False)
        self.check_is_bool("include_punctuation")
        self.max_unique_query_attempts = kwargs.pop(
            "max_unique_query_attempts", MAX_UNIQUE_QUERY_ATTEMPTS
        )

        # Set unique=False unless it's been set manually.
        if "unique" not in kwargs:
            kwargs["unique"] = False

        super().__init__(*args, **kwargs)

    def random_char_generator(self, chars):
        for i in range(self.max_unique_query_attempts):
            yield "".join(get_random_string(self.length, chars))
        raise RuntimeError(
            "max random character attempts exceeded (%s)"
            % self.max_unique_query_attempts
        )

    def in_unique_together(self, model_instance):
        for params in model_instance._meta.unique_together:
            if self.attname in params:
                return True
        return False

    def pre_save(self, model_instance, add):
        if (not add or self.keep_default) and getattr(
            model_instance, self.attname
        ) != "":
            return getattr(model_instance, self.attname)

        population = ""
        if self.include_alpha:
            if self.lowercase:
                population += string.ascii_lowercase
            elif self.uppercase:
                population += string.ascii_uppercase
            else:
                population += string.ascii_letters

        if self.include_digits:
            population += string.digits

        if self.include_punctuation:
            population += string.punctuation

        random_chars = self.random_char_generator(population)
        if not self.unique and not self.in_unique_together(model_instance):
            new = next(random_chars)
            setattr(model_instance, self.attname, new)
            return new

        return self.find_unique(
            model_instance,
            model_instance._meta.get_field(self.attname),
            random_chars,
        )

    def internal_type(self):
        return "CharField"

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["length"] = self.length
        del kwargs["max_length"]
        if self.lowercase is True:
            kwargs["lowercase"] = self.lowercase
        if self.uppercase is True:
            kwargs["uppercase"] = self.uppercase
        if self.include_alpha is False:
            kwargs["include_alpha"] = self.include_alpha
        if self.include_digits is False:
            kwargs["include_digits"] = self.include_digits
        if self.include_punctuation is True:
            kwargs["include_punctuation"] = self.include_punctuation
        if self.unique is True:
            kwargs["unique"] = self.unique
        return name, path, args, kwargs


class CreationDateTimeField(DateTimeField):
    """
    CreationDateTimeField

    By default, sets editable=False, blank=True, auto_now_add=True
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("editable", False)
        kwargs.setdefault("blank", True)
        kwargs.setdefault("auto_now_add", True)
        DateTimeField.__init__(self, *args, **kwargs)

    def get_internal_type(self):
        return "DateTimeField"

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if self.editable is not False:
            kwargs["editable"] = True
        if self.blank is not True:
            kwargs["blank"] = False
        if self.auto_now_add is not False:
            kwargs["auto_now_add"] = True
        return name, path, args, kwargs


class ModificationDateTimeField(CreationDateTimeField):
    """
    ModificationDateTimeField

    By default, sets editable=False, blank=True, auto_now=True

    Sets value to now every time the object is saved.
    """

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("auto_now", True)
        DateTimeField.__init__(self, *args, **kwargs)

    def get_internal_type(self):
        return "DateTimeField"

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if self.auto_now is not False:
            kwargs["auto_now"] = True
        return name, path, args, kwargs

    def pre_save(self, model_instance, add):
        if not getattr(model_instance, "update_modified", True):
            return getattr(model_instance, self.attname)
        return super().pre_save(model_instance, add)


class UUIDVersionError(Exception):
    pass


class UUIDFieldMixin:
    """
    UUIDFieldMixin

    By default uses UUID version 4 (randomly generated UUID).

    The field support all uuid versions which are natively supported by the uuid python module, except version 2.
    For more information see: https://docs.python.org/lib/module-uuid.html
    """  # noqa: E501

    DEFAULT_MAX_LENGTH = 36

    def __init__(
        self,
        verbose_name=None,
        name=None,
        auto=True,
        version=4,
        node=None,
        clock_seq=None,
        namespace=None,
        uuid_name=None,
        *args,
        **kwargs,
    ):
        if not HAS_UUID:
            raise ImproperlyConfigured(
                "'uuid' module is required for UUIDField. "
                "(Do you have Python 2.5 or higher installed ?)"
            )

        kwargs.setdefault("max_length", self.DEFAULT_MAX_LENGTH)

        if auto:
            self.empty_strings_allowed = False
            kwargs["blank"] = True
            kwargs.setdefault("editable", False)

        self.auto = auto
        self.version = version
        self.node = node
        self.clock_seq = clock_seq
        self.namespace = namespace
        self.uuid_name = uuid_name or name

        super().__init__(verbose_name=verbose_name, *args, **kwargs)

    def create_uuid(self):
        if not self.version or self.version == 4:
            return uuid.uuid4()
        elif self.version == 1:
            return uuid.uuid1(self.node, self.clock_seq)
        elif self.version == 2:
            raise UUIDVersionError("UUID version 2 is not supported.")
        elif self.version == 3:
            return uuid.uuid3(self.namespace, self.uuid_name)
        elif self.version == 5:
            return uuid.uuid5(self.namespace, self.uuid_name)
        else:
            raise UUIDVersionError("UUID version %s is not valid." % self.version)

    def pre_save(self, model_instance, add):
        value = super().pre_save(model_instance, add)

        if self.auto and add and value is None:
            value = force_str(self.create_uuid())
            setattr(model_instance, self.attname, value)
            return value
        else:
            if self.auto and not value:
                value = force_str(self.create_uuid())
                setattr(model_instance, self.attname, value)

        return value

    def formfield(self, form_class=None, choices_form_class=None, **kwargs):
        if self.auto:
            return None
        return super().formfield(form_class, choices_form_class, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()

        if kwargs.get("max_length", None) == self.DEFAULT_MAX_LENGTH:
            del kwargs["max_length"]
        if self.auto is not True:
            kwargs["auto"] = self.auto
        if self.version != 4:
            kwargs["version"] = self.version
        if self.node is not None:
            kwargs["node"] = self.node
        if self.clock_seq is not None:
            kwargs["clock_seq"] = self.clock_seq
        if self.namespace is not None:
            kwargs["namespace"] = self.namespace
        if self.uuid_name is not None:
            kwargs["uuid_name"] = self.name

        return name, path, args, kwargs


class ShortUUIDField(UUIDFieldMixin, CharField):
    """
    ShortUUIDFied

    Generates concise (22 characters instead of 36), unambiguous, URL-safe UUIDs.

    Based on `shortuuid`: https://github.com/stochastic-technologies/shortuuid
    """

    DEFAULT_MAX_LENGTH = 22

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not HAS_SHORT_UUID:
            raise ImproperlyConfigured(
                "'shortuuid' module is required for ShortUUIDField. "
                "(Do you have Python 2.5 or higher installed ?)"
            )
        kwargs.setdefault("max_length", self.DEFAULT_MAX_LENGTH)

    def create_uuid(self):
        if not self.version or self.version == 4:
            return shortuuid.uuid()
        elif self.version == 1:
            return shortuuid.uuid()
        elif self.version == 2:
            raise UUIDVersionError("UUID version 2 is not supported.")
        elif self.version == 3:
            raise UUIDVersionError("UUID version 3 is not supported.")
        elif self.version == 5:
            return shortuuid.uuid(name=self.namespace)
        else:
            raise UUIDVersionError("UUID version %s is not valid." % self.version)



# -*- coding: utf-8 -*-
import fnmatch
import os
import py_compile
from os.path import join as _j
from typing import List

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from django_extensions.management.utils import signalcommand


class Command(BaseCommand):
    help = "Compile python bytecode files for the project."
    requires_system_checks: List[str] = []

    def add_arguments(self, parser):
        parser.add_argument(
            "--path",
            "-p",
            action="store",
            dest="path",
            help="Specify path to recurse into",
        )

    @signalcommand
    def handle(self, *args, **options):
        project_root = options["path"]
        if not project_root:
            project_root = getattr(settings, "BASE_DIR", None)

        verbosity = options["verbosity"]
        if not project_root:
            raise CommandError(
                "No --path specified and settings.py does not contain BASE_DIR"
            )

        for root, dirs, filenames in os.walk(project_root):
            for filename in fnmatch.filter(filenames, "*.py"):
                full_path = _j(root, filename)
                if verbosity > 1:
                    self.stdout.write("Compiling %s...\n" % full_path)
                py_compile.compile(full_path)



# -*- coding: utf-8 -*-
"""
JSONField automatically serializes most Python terms to JSON data.
Creates a TEXT field with a default value of "{}".  See test_json.py for
more information.

 from django.db import models
 from django_extensions.db.fields import json

 class LOL(models.Model):
     extra = json.JSONField()
"""

import json

from django.core.serializers.json import DjangoJSONEncoder
from django.db import models
from django.db.models import expressions


def dumps(value):
    return DjangoJSONEncoder().encode(value)


def loads(txt):
    return json.loads(txt)


class JSONDict(dict):
    """
    Hack so repr() called by dumpdata will output JSON instead of
    Python formatted data.  This way fixtures will work!
    """

    def __repr__(self):
        return dumps(self)


class JSONList(list):
    """
    Hack so repr() called by dumpdata will output JSON instead of
    Python formatted data.  This way fixtures will work!
    """

    def __repr__(self):
        return dumps(self)


class JSONField(models.TextField):
    """
    JSONField is a generic textfield that neatly serializes/unserializes
    JSON objects seamlessly.  Main thingy must be a dict object.
    """

    def __init__(self, *args, **kwargs):
        kwargs["default"] = kwargs.get("default", dict)
        models.TextField.__init__(self, *args, **kwargs)

    def get_default(self):
        if self.has_default():
            default = self.default

            if callable(default):
                default = default()

            return self.to_python(default)
        return super().get_default()

    def to_python(self, value):
        """Convert our string value to JSON after we load it from the DB"""
        if value is None or value == "":
            return {}

        if isinstance(value, str):
            res = loads(value)
        else:
            res = value

        if isinstance(res, dict):
            return JSONDict(**res)
        elif isinstance(res, list):
            return JSONList(res)

        return res

    def get_prep_value(self, value):
        if not isinstance(value, str):
            return dumps(value)
        return super(models.TextField, self).get_prep_value(value)

    def from_db_value(self, value, expression, connection):  # type: ignore
        return self.to_python(value)

    def get_db_prep_save(self, value, connection, **kwargs):
        """Convert our JSON object to a string before we save"""
        if value is None and self.null:
            return None

        # default values come in as strings; only non-strings should be
        # run through `dumps`
        if (
            not isinstance(value, str)
            # https://github.com/django-extensions/django-extensions/issues/1924
            # https://code.djangoproject.com/ticket/35167
            and not isinstance(value, expressions.Expression)
        ):
            value = dumps(value)

        return super().get_db_prep_save(value, connection)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if self.default == "{}":
            del kwargs["default"]
        return name, path, args, kwargs
